# EBML/Matroska parser
# Copyright (C) 2010, 2015  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


# This code is heavily based on public domain code by "Omion" (from the
# Hydrogenaudio forums), as obtained from Matroska's Subversion repository at
# revision 858 (2004-10-03), under "/trunk/Perl.Parser/MatroskaParser.pm".


import sys
from struct import pack, unpack
from warnings import warn

SINT, UINT, FLOAT, STRING, UTF8, DATE, MASTER, BINARY = range(8)


class EbmlException(Exception):
    pass


class EbmlWarning(Warning):
    pass


class BinaryData(bytes):
    def __repr__(self):
        return "<BinaryData>"


def bchr(n):
    """chr() that always returns bytes in Python 2 and 3"""
    return pack('B', n)


class Ebml:
    """EBML parser.

    Usage: Ebml(location, tags).parse()
    where `tags` is a dictionary of the form {id: (name, type)}.
    """

    ## Constructor and destructor

    def __init__(self, location, tags):
        self.tags = tags
        self.open(location)

    def __del__(self):
        self.close()

    ## File access.
    ## These can be overridden to provide network support.

    def open(self, location):
        """Open a location and set self.size."""
        self.file = f = open(location, 'rb')
        f = self.file
        f.seek(0, 2)
        self.size = f.tell()
        f.seek(0, 0)

    def seek(self, offset, mode):
        self.file.seek(offset, mode)

    def tell(self):
        return self.file.tell()

    def read(self, length):
        return self.file.read(length)

    def close(self):
        self.file.close()

    ## Element reading

    def readID(self):
        b = self.read(1)
        b1 = ord(b)
        if b1 & 0b10000000:  # 1 byte
            return b
        elif b1 & 0b01000000:  # 2 bytes
            return unpack(">H", b + self.read(1))[0]
        elif b1 & 0b00100000:  # 3 bytes
            return unpack(">L", b"\0" + b + self.read(2))[0]
        elif b1 & 0b00010000:  # 4 bytes
            return unpack(">L", b + self.read(3))[0]
        else:
            raise EbmlException("invalid element ID (leading byte 0x%02X)" % b1)

    def readSize(self):
        b1 = ord(self.read(1))
        if b1 & 0b10000000:  # 1 byte
            return b1 & 0b01111111
        elif b1 & 0b01000000:  # 2 bytes
            return unpack(">H", bchr(b1 & 0b00111111) + self.read(1))[0]
        elif b1 & 0b00100000:  # 3 bytes
            return unpack(">L", b"\0" + bchr(b1 & 0b00011111) + self.read(2))[0]
        elif b1 & 0b00010000:  # 4 bytes
            return unpack(">L", bchr(b1 & 0b00001111) + self.read(3))[0]
        elif b1 & 0x00001000:  # 5 bytes
            return unpack(">Q", b"\0\0\0" + bchr(b1 & 0b00000111) + self.read(4))[0]
        elif b1 & 0b00000100:  # 6 bytes
            return unpack(">Q", b"\0\0" + bchr(b1 & 0b0000011) + self.read(5))[0]
        elif b1 & 0b00000010:  # 7 bytes
            return unpack(">Q", b"\0" + bchr(b1 & 0b00000001) + self.read(6))[0]
        elif b1 & 0b00000001:  # 8 bytes
            return unpack(">Q", b"\0" + self.read(7))[0]
        else:
            assert b1 == 0
            raise EbmlException("undefined element size")

    def readInteger(self, length, signed):
        if length == 1:
            value = ord(self.read(1))
        elif length == 2:
            value = unpack(">H", self.read(2))[0]
        elif length == 3:
            value = unpack(">L", b"\0" + self.read(3))[0]
        elif length == 4:
            value = unpack(">L", self.read(4))[0]
        elif length == 5:
            value = unpack(">Q", b"\0\0\0" + self.read(5))[0]
        elif length == 6:
            value = unpack(">Q", b"\0\0" + self.read(6))[0]
        elif length == 7:
            value = unpack(">Q", b"\0" + (self.read(7)))[0]
        elif length == 8:
            value = unpack(">Q", self.read(8))[0]
        else:
            raise EbmlException("don't know how to read %r-byte integer" % length)
        if signed:
            nbits = (8 - length) + 8 * (length - 1)
            if value >= (1 << (nbits - 1)):
                value -= 1 << nbits
        return value

    def readFloat(self, length):
        if length == 4:
            return unpack('>f', self.read(4))[0]
        elif length == 8:
            return unpack('>d', self.read(8))[0]
        else:
            raise EbmlException("don't know how to read %r-byte float" % length)

    ## Parsing

    def parse(self, from_=0, to=None):
        """Parses EBML from `from_` (inclusive) to `to` (exclusive).

        Note that not all streams support seeking backwards, so prepare to handle
        an exception if you try to parse from arbitrary position.
        """
        if to is None:
            to = self.size
        self.seek(from_, 0)
        node = {}
        # Iterate over current node's children.
        while self.tell() < to:
            try:
                id = self.readID()
            except EbmlException as e:
                # Invalid EBML header. We can't reliably get any more data from
                # this level, so just return anything we have.
                warn(EbmlWarning(e))
                return node
            size = self.readSize()
            if size == 0b01111111:
                warn(EbmlWarning("don't know how to handle unknown-sized element"))
                size = to - self.tell()
            try:
                key, type_ = self.tags[id]
            except KeyError:
                self.seek(size, 1)
                continue
            try:
                if type_ is SINT:
                    value = self.readInteger(size, True)
                elif type_ is UINT:
                    value = self.readInteger(size, False)
                elif type_ is FLOAT:
                    value = self.readFloat(size)
                elif type_ is STRING:
                    value = self.read(size).decode('ascii')
                elif type_ is UTF8:
                    value = self.read(size).decode('utf-8')
                elif type_ is DATE:
                    us = self.readInteger(size, True) / 1000.0  # ns to us
                    from datetime import datetime, timedelta

                    value = datetime(2001, 1, 1) + timedelta(microseconds=us)
                elif type_ is MASTER:
                    tell = self.tell()
                    value = self.parse(tell, tell + size)
                elif type_ is BINARY:
                    value = BinaryData(self.read(size))
                else:
                    assert False, type_
            except (EbmlException, UnicodeDecodeError) as e:
                warn(EbmlWarning(e))
            else:
                try:
                    parentval = node[key]
                except KeyError:
                    parentval = node[key] = []
                parentval.append(value)
        return node


## GIO-specific code

from gi.repository import Gio


class GioEbml(Ebml):
    # NOTE: All seeks are faked using InputStream.skip because we need to use
    # BufferedInputStream but it does not implement Seekable.

    def open(self, location):
        f = Gio.File.new_for_commandline_arg(location)
        self.buffer = Gio.BufferedInputStream.new(f.read())
        self._tell = 0

        self.size = f.query_info(
            'standard::size', Gio.FileQueryInfoFlags.NONE, None
        ).get_size()

    def seek(self, offset, mode):
        if mode == 0:
            skip = offset - self._tell
        elif mode == 1:
            skip = offset
        elif mode == 2:
            skip = self.size - self._tell + offset
        else:
            raise ValueError("invalid seek mode: %r" % mode)
        if skip < 0:
            raise NotImplementedError("seeking backwards not supported")
        self._tell += skip
        self.buffer.skip(skip)

    def tell(self):
        return self._tell

    def read(self, length):
        result = self.buffer.read_bytes(length).get_data()
        self._tell += len(result)
        return result

    def close(self):
        self.buffer.close()


## Matroska-specific code

# Interesting Matroska tags.
# Tags not defined here are skipped while parsing.
MatroskaTags = {
    # Segment
    0x18538067: ('Segment', MASTER),
    # Segment Information
    0x1549A966: ('Info', MASTER),
    0x7384: ('SegmentFilename', UTF8),
    0x2AD7B1: ('TimecodeScale', UINT),
    0x4489: ('Duration', FLOAT),
    0x4461: ('DateUTC', DATE),
    0x7BA9: ('Title', UTF8),
    0x4D80: ('MuxingApp', UTF8),
    0x5741: ('WritingApp', UTF8),
    # Track
    0x1654AE6B: ('Tracks', MASTER),
    0xAE: ('TrackEntry', MASTER),
    0xD7: ('TrackNumber', UINT),
    0x83: ('TrackType', UINT),
    0xB9: ('FlagEnabled', UINT),
    0x88: ('FlagDefault', UINT),
    0x23E383: ('DefaultDuration', UINT),
    0x536E: ('Name', UTF8),
    0x22B59C: ('Language', STRING),
    0x86: ('CodecID', STRING),
    0x258688: ('CodecName', UTF8),
    # Video
    0xE0: ('Video', MASTER),
    # Audio
    0xE1: ('Audio', MASTER),
    0xB5: ('SamplingFrequency', FLOAT),
    0x78B5: ('OutputSamplingFrequency', FLOAT),
    0x9F: ('Channels', UINT),
    0x6264: ('BitDepth', UINT),
    # Attachment
    0x1941A469: ('Attachments', MASTER),
    0x61A7: ('AttachedFile', MASTER),
    0x466E: ('FileName', UTF8),
    0x465C: ('FileData', BINARY),
    # Chapters
    0x1043A770: ('Chapters', MASTER),
    0x45B9: ('EditionEntry', MASTER),
    0x45BC: ('EditionUID', UINT),
    0x45BD: ('EditionFlagHidden', UINT),
    0x45DB: ('EditionFlagDefault', UINT),
    0x45DD: ('EditionFlagOrdered', UINT),
    0xB6: ('ChapterAtom', MASTER),
    0x73C4: ('ChapterUID', UINT),
    0x91: ('ChapterTimeStart', UINT),
    0x92: ('ChapterTimeEnd', UINT),
    0x98: ('ChapterFlagHidden', UINT),
    0x4598: ('ChapterFlagEnabled', UINT),
    0x63C3: ('ChapterPhysicalEquiv', UINT),
    0x8F: ('ChapterTrack', MASTER),
    0x89: ('ChapterTrackNumber', UINT),
    0x80: ('ChapterDisplay', MASTER),
    0x85: ('ChapString', UTF8),
    0x437C: ('ChapLanguage', STRING),
    0x437E: ('ChapCountry', STRING),
    # Tagging
    0x1254C367: ('Tags', MASTER),
    0x7373: ('Tag', MASTER),
    0x63C0: ('Targets', MASTER),
    0x68CA: ('TargetTypevalue', UINT),
    0x63CA: ('TargetType', STRING),
    0x63C5: ('TagTrackUID', UINT),
    0x63C9: ('TagEditionUID', UINT),
    0x63C4: ('TagChapterUID', UINT),
    0x67C8: ('SimpleTag', MASTER),
    0x45A3: ('TagName', UTF8),
    0x447A: ('TagLanguage', STRING),
    0x4484: ('TagDefault', UINT),
    0x4487: ('TagString', UTF8),
    0x4485: ('TagBinary', BINARY),
}


def parse(location):
    return GioEbml(location, MatroskaTags).parse()


def dump(location):
    from pprint import pprint

    pprint(parse(location))


def dump_tags(location):
    from pprint import pprint

    mka = parse(location)
    segment = mka['Segment'][0]
    info = segment['Info'][0]
    try:
        timecodescale = info['TimecodeScale'][0]
    except KeyError:
        timecodescale = 1000000
    length = info['Duration'][0] * timecodescale / 1e9
    print("Length = %s seconds" % length)
    pprint(segment['Tags'][0]['Tag'])


def gio_location(location):
    """Convert location to GIO-compatible location.

    This works around broken behaviour in the Win32 GIO port (it converts paths
    into UTF-8 and requires them to be specified in UTF-8 as well).

    :type location: str
    :rtype: bytes
    """
    if sys.platform == 'win32' and '://' not in location:
        if isinstance(location, bytes):
            # Decode the path according to the FS encoding to get the Unicode
            # representation first. If the path is in a different encoding,
            # this step will fail.
            location = location.decode(sys.getfilesystemencoding())
        location = location.encode('utf-8')
    return location


if __name__ == '__main__':
    location = gio_location(sys.argv[1])
    dump_tags(location)


# vi: et sts=4 sw=4 ts=4



# -*- coding: utf-8 -*-
# Copyright (C) 2010 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Provides an extensible framework for processing and
preparation of data for display in various contexts.
"""

from datetime import date
from gi.repository import GLib
from gi.repository import GObject
import re

from xl import common, providers, settings, trax
from xl.common import TimeSpan
from xl.nls import gettext as _, ngettext


# NOTE: the following two classes used to subclass string._TemplateMetaclass
# and string.Template from string module. However, python 3.9 reorganized
# the string.Template class and removed the string._TemplateMetaclass. It
# also turned out that our implementation's overridden functionality is
# self-sufficient, so these are stand-alone classes now (and contain
# partial copy of the corresponding base classes from python 3.8 and earlier).
class _ParameterTemplateMetaclass(type):
    # Allows for $tag, ${tag}, ${tag:parameter} and ${tag:parameter=argument}
    pattern = r"""
    %(delim)s(?:
      (?P<escaped>%(delim)s) |     # Escape sequence of two delimiters
      (?P<named>%(id)s)      |     # Delimiter and a Python identifier
      {
        (?P<braced>%(id)s)         # Delimiter and a braced identifier
        (!?:
          (?P<parameters>
            %(id)s                 # First optional parameter indicated with ':'
            (!?=(!?%(arg)s)+?)?    # Optional argument indicated with '='
            (!?,\s*%(id)s          # Further optional parameters separated with ','
              (!?=(!?%(arg)s)+?)?  # Optional argument indicated with '='
            )*
          )
        )?
      }                      |
      (?P<invalid>)                # Other ill-formed delimiter expressions
    )
    """

    def __init__(cls, name, bases, dct):
        super(_ParameterTemplateMetaclass, cls).__init__(name, bases, dct)
        if 'pattern' in dct:
            pattern = cls.pattern
        else:
            pattern = _ParameterTemplateMetaclass.pattern % {
                'delim': re.escape(cls.delimiter),
                'id': cls.idpattern,
                'arg': cls.argpattern,
            }
        cls.pattern = re.compile(pattern, re.IGNORECASE | re.VERBOSE)


class ParameterTemplate(metaclass=_ParameterTemplateMetaclass):
    """
    An extended template class which additionally
    accepts parameters assigned to identifiers.

    This introduces another pattern group named
    "parameters" in addition to the groups
    created by :class:`string.Template`

    Examples:

    * ``${foo:parameter1}``
    * ``${bar:parameter1, parameter2}``
    * ``${qux:parameter1=argument1, parameter2}``
    """

    delimiter = '$'
    idpattern = r'[_a-z][_a-z0-9]*'
    argpattern = r'[^,}=]|\,|\}|\='

    def __init__(self, template):
        """
        :param template: the template string
        """
        self.template = template

    def safe_substitute(self, *args):
        """
        Overridden to allow for parametrized identifiers
        """
        if len(args) > 1:
            raise TypeError('Too many positional arguments')

        if not args:
            mapping = {}
        else:
            mapping = args[0]

        # Helper function for .sub()
        def convert(mo):
            named = mo.group('named')

            if named is not None:
                try:
                    # We use this idiom instead of str() because the latter
                    # will fail if val is a Unicode containing non-ASCII
                    return '%s' % (mapping[named],)
                except KeyError:
                    return self.delimiter + named

            braced = mo.group('braced')

            if braced is not None:
                parts = [braced]
                parameters = mo.group('parameters')

                if parameters is not None:
                    parts += [parameters]

                try:
                    return '%s' % (mapping[':'.join(parts)],)
                except KeyError:
                    return self.delimiter + '{' + ':'.join(parts) + '}'

            if mo.group('escaped') is not None:
                return self.delimiter

            if mo.group('invalid') is not None:
                return self.delimiter

            raise ValueError('Unrecognized named group in pattern', self.pattern)

        return self.pattern.sub(convert, self.template)


class Formatter(GObject.GObject):
    """
    A generic text formatter based on a format string

    By default the following parameters are provided
    to each identifier:

    * ``prefix``, ``suffix``: a string to put before or after the formatted string if that string is not empty
        * Whitespace will be not be touched and transferred as is
        * The characters ``,``, ``}`` and ``=`` need to be escaped like ``\,``, ``\}`` and ``\=`` respectively
    * ``pad``: desired length the formatted string should have, will be achieved using the ``padstring``
    * ``padstring``: a string to use for padding, will be repeated as often as possible to achieve the desired length specified by ``pad``
        * Example: ``${identifier:pad=4, padstring=XY}`` for *identifier* having the value *a* will become *XYXa*
    """

    __gproperties__ = {
        'format': (
            GObject.TYPE_STRING,
            'format string',
            'String the formatting is based on',
            '',
            GObject.ParamFlags.READWRITE,
        )
    }

    def __init__(self, format):
        """
        :param format: the initial format, see the documentation
            of :class:`string.Template` for details
        :type format: string
        """
        GObject.GObject.__init__(self)

        self._template = ParameterTemplate(format)
        self._substitutions = {}

    def do_get_property(self, property):
        """
        Gets GObject properties
        """
        if property.name == 'format':
            return self._template.template
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Sets GObject properties
        """
        if property.name == 'format':
            if value != self._template.template:
                self._template.template = value
        else:
            raise AttributeError('unknown property %s' % property.name)

    def extract(self):
        """
        Retrieves the identifiers and their optional parameters

        Example of the returned dictionary::

            extractions = {
                'identifier1': (
                    'identifier1', {}),
                'identifier2:parameter': (
                    'identifier2', {'parameter': True}),
                'identifier3:parameter=argument': (
                    'identifier3', {'parameter': 'argument'})
            }

        :returns: the extractions
        :rtype: dict
        """
        matches = self._template.pattern.finditer(self._template.template)
        extractions = {}

        # Extract list of identifiers and parameters from the format string
        for match in matches:
            groups = match.groupdict()

            # We only care about braced and named, not escaped and invalid
            identifier = groups['braced'] or groups['named']

            if identifier is None:
                continue

            identifier_parts = [identifier]
            parameters = {}

            if groups['parameters'] is not None:
                # Split parameters on unescaped comma
                parameters = [
                    p.lstrip() for p in re.split(r'(?<!\\),', groups['parameters'])
                ]
                # Split arguments on unescaped equals sign
                parameters = [
                    (re.split(r'(?<!\\)=', p, 1) + [True])[:2] for p in parameters
                ]
                # Turn list of lists into a proper dictionary
                parameters = dict(parameters)

                # Remove now obsolete escapes
                for p in parameters:
                    argument = parameters[p]

                    if not isinstance(argument, bool):
                        argument = argument.replace(r'\,', ',')
                        argument = argument.replace(r'\}', '}')
                        argument = argument.replace(r'\=', '=')
                        parameters[p] = argument

                identifier_parts += [groups['parameters']]

            # Required to make multiple occurences of the same
            # identifier with different parameters work
            extractions[':'.join(identifier_parts)] = (identifier, parameters)

        return extractions

    def format(self, *args):
        """
        Returns a string by formatting the passed data

        :param args: data to base the formatting on
        :returns: the formatted text
        :rtype: string
        """
        extractions = self.extract()
        substitutions = {}

        for needle, (identifier, parameters) in extractions.items():
            substitute = None

            if needle in self._substitutions:
                substitute = self._substitutions[needle]
            elif identifier in self._substitutions:
                substitute = self._substitutions[identifier]

            if substitute is not None:
                prefix = parameters.pop('prefix', '')
                suffix = parameters.pop('suffix', '')
                pad = int(parameters.pop('pad', 0))
                padstring = parameters.pop('padstring', '')

                if callable(substitute):
                    substitute = substitute(*args, **parameters)

                if pad > 0 and padstring:
                    # Decrease pad length by value length
                    pad = max(0, pad - len(substitute))
                    # Retrieve the maximum multiplier for the pad string
                    padcount = pad // len(padstring) + 1
                    # Generate pad string
                    padstring = padcount * padstring
                    # Clamp pad string
                    padstring = padstring[0:pad]
                    substitute = '%s%s' % (padstring, substitute)

                if substitute:
                    substitute = '%s%s%s' % (prefix, substitute, suffix)

                substitutions[needle] = substitute

        return self._template.safe_substitute(substitutions)


class ProgressTextFormatter(Formatter):
    """
    A text formatter for progress indicators
    """

    def __init__(self, format, player):
        Formatter.__init__(self, format)
        self._player = player

    def format(self, current_time=None, total_time=None):
        """
        Returns a string suitable for progress indicators

        :param current_time: the current progress, taken from the current playback if not set
        :type current_time: float
        :param total_time: the total length of a track, taken from the current playback if not set
        :type total_time: float
        :returns: The formatted text
        :rtype: string
        """
        total_remaining_time = 0

        if current_time is None:
            current_time = self._player.get_time()

        if total_time is None:
            track = self._player.current

            if track is not None:
                total_time = track.get_tag_raw('__length')

        if total_time is None:
            total_time = remaining_time = 0
        else:
            remaining_time = total_time - current_time

        playlist = self._player.queue.current_playlist

        if playlist and playlist.current_position >= 0:
            tracks = playlist[playlist.current_position :]
            total_remaining_time = sum(t.get_tag_raw('__length') or 0 for t in tracks)
            total_remaining_time -= current_time

        self._substitutions['current_time'] = LengthTagFormatter.format_value(
            current_time
        )
        self._substitutions['remaining_time'] = LengthTagFormatter.format_value(
            remaining_time
        )
        self._substitutions['total_time'] = LengthTagFormatter.format_value(total_time)
        self._substitutions['total_remaining_time'] = LengthTagFormatter.format_value(
            total_remaining_time
        )

        return Formatter.format(self)


class TrackFormatter(Formatter):
    """
    A formatter for track data
    """

    def format(self, track, markup_escape=False):
        """
        Returns a string for places where
        track data is presented to the user

        :param track: a single track to take data from
        :type track: :class:`xl.trax.Track`
        :param markup_escape: whether to escape markup-like
            characters in tag values
        :type markup_escape: bool
        :returns: the formatted text
        :rtype: string
        """
        if not isinstance(track, trax.Track):
            raise TypeError(
                'First argument to format() needs ' 'to be of type xl.trax.Track'
            )

        extractions = self.extract()
        self._substitutions = {}

        for identifier, (tag, parameters) in extractions.items():
            provider = providers.get_provider('tag-formatting', tag)

            if provider is None:
                substitute = track.get_tag_display(tag)
            else:
                substitute = provider.format(track, parameters)

            if markup_escape:
                substitute = GLib.markup_escape_text(substitute)

            self._substitutions[identifier] = substitute

        return Formatter.format(self)


class TagFormatter:
    """
    A formatter provider for a tag of a track
    """

    def __init__(self, name):
        """
        :param name: the name of the tag
        :type name: string
        """
        self.name = name

    def format(self, track, parameters):
        """
        Formats a raw tag value. Accepts optional
        parameters to manipulate the formatting
        process.

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        pass


class NumberTagFormatter(TagFormatter):
    """
    A generic formatter for numeric formatting

    Removes count values, e.g. "b" in "a/b"
    """

    def __init__(self, name):
        """
        :param name: the name of the tag
        :type name: string
        """
        TagFormatter.__init__(self, name)

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        value = track.get_tag_raw(self.name, join=True)

        if not value:
            return ''

        try:  # n/n
            value, count = value.split('/')
        except ValueError:  # n
            pass

        if not value:
            return ''

        try:
            value = int(value)
        except ValueError:
            return ''

        return '%d' % value


class TrackNumberTagFormatter(NumberTagFormatter):
    """
    A formatter for the tracknumber of a track
    """

    def __init__(self):
        TagFormatter.__init__(self, 'tracknumber')


providers.register('tag-formatting', TrackNumberTagFormatter())


class DiscNumberTagFormatter(NumberTagFormatter):
    """
    A formatter for the discnumber of a track
    """

    def __init__(self):
        TagFormatter.__init__(self, 'discnumber')


providers.register('tag-formatting', DiscNumberTagFormatter())


class ArtistTagFormatter(TagFormatter):
    """
    A formatter for the artist of a track
    """

    def __init__(self):
        TagFormatter.__init__(self, 'artist')

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
            Possible values are:

            * compilate:
              Allows for proper handling of compilations,
              either via albumartist tag, a fallback value,
              or simply all artists
        :returns: the formatted value
        :rtype: string
        """
        compilate = parameters.get('compilate', False)
        value = track.get_tag_display(self.name, artist_compilations=compilate)

        return value


providers.register('tag-formatting', ArtistTagFormatter())


class TimeTagFormatter(TagFormatter):
    """
    A formatter for a time period
    """

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: Verbosity of the output,
            possible values for "format" are:

            * short: "1:02:42"
            * long: "1h, 2m, 42s"
            * verbose: "1 hour, 2 minutes, 42 seconds"
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        value = track.get_tag_raw(self.name)
        format = parameters.get('format', 'short')

        return self.format_value(value, format)

    @staticmethod
    def format_value(value, format='short'):
        """
        Formats a length value

        :param value: the length in seconds
        :type value: float
        :param format: verbosity of the output,
            possible values are:

            * short: "1:02:42"
            * long: "1h, 2m, 42s"
            * verbose: "1 hour, 2 minutes, 42 seconds"
        :type format: string
        :returns: the formatted value
        :rtype: string
        """
        span = TimeSpan(value)
        text = ''

        if format == 'verbose':
            if span.days > 0:
                text += ngettext('%d day, ', '%d days, ', span.days) % span.days
            if span.hours > 0:
                text += ngettext('%d hour, ', '%d hours, ', span.hours) % span.hours
            text += ngettext('%d minute, ', '%d minutes, ', span.minutes) % span.minutes
            text += ngettext('%d second', '%d seconds', span.seconds) % span.seconds

        elif format == 'long':
            if span.days > 0:
                # TRANSLATORS: Short form of an amount of days
                text += _('%dd, ') % span.days
            if span.hours > 0:
                # TRANSLATORS: Short form of an amount of hours
                text += _('%dh, ') % span.hours
            # TRANSLATORS: Short form of an amount of minutes
            text += _('%dm, ') % span.minutes
            # TRANSLATORS: Short form of an amount of seconds
            text += _('%ds') % span.seconds

        elif format == 'short':
            if span.days > 0:
                # TRANSLATORS: Short form of an amount of days
                text += _('%dd ') % span.days
            if span.hours > 0 or text:  # always show hours when > 1 day
                # TRANSLATORS: Time duration (hours:minutes:seconds)
                text += _('%d:%02d:%02d') % (span.hours, span.minutes, span.seconds)
            else:
                # TRANSLATORS: Time duration (minutes:seconds)
                text += _('%d:%02d') % (span.minutes, span.seconds)

        else:
            raise ValueError(
                'Invalid argument "%s" passed to parameter '
                '"format" for tag "__length", possible arguments are '
                '"short", "long" and "verbose"' % format
            )

        return text


class LengthTagFormatter(TimeTagFormatter):
    """
    A formatter for the length of a track
    """

    def __init__(self):
        TimeTagFormatter.__init__(self, '__length')


providers.register('tag-formatting', LengthTagFormatter())


class StartOffsetTagFormatter(TimeTagFormatter):
    """
    A formatter for the track's start offset
    """

    def __init__(self):
        TimeTagFormatter.__init__(self, '__startoffset')


providers.register('tag-formatting', StartOffsetTagFormatter())


class StopOffsetTagFormatter(TimeTagFormatter):
    """
    A formatter for the track's stop offset
    """

    def __init__(self):
        TimeTagFormatter.__init__(self, '__stopoffset')


providers.register('tag-formatting', StopOffsetTagFormatter())


class RatingTagFormatter(TagFormatter):
    """
    A formatter for the rating of a track

    Will return glyphs representing the rating like ★★★☆☆
    """

    def __init__(self):
        TagFormatter.__init__(self, '__rating')

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        rating = track.get_rating()
        maximum = settings.get_option('rating/maximum', 5)

        filled = '★' * int(rating)
        empty = '☆' * int(maximum - rating)
        return filled + empty


providers.register('tag-formatting', RatingTagFormatter())


class YearTagFormatter(TagFormatter):
    """
    A pseudo-tag that computes the year from the date column
    """

    def __init__(self):
        TagFormatter.__init__(self, 'year')

    def format(self, track, parameters):
        value = track.get_tag_raw('date')
        if value is not None:
            try:
                return value[0].split('-')[0]
            except Exception:
                pass
            return value[0]

        return _("Unknown")


providers.register('tag-formatting', YearTagFormatter())


class DateTagFormatter(TagFormatter):
    """
    A generic formatter for timestamp formatting

    Will return the localized string for *Today*, *Yesterday*
    or the respective localized date for earlier dates
    """

    def __init__(self, name):
        """
        :param name: the name of the tag
        :type name: string
        """
        TagFormatter.__init__(self, name)

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        value = track.get_tag_raw(self.name)
        # TRANSLATORS: Indicates that a track has never been played before
        text = _('Never')

        try:
            last_played = date.fromtimestamp(value)
        except (TypeError, ValueError):
            text = _('Never')
        else:
            today = date.today()
            delta = today - last_played

            if delta.days == 0:
                text = _('Today')
            elif delta.days == 1:
                text = _('Yesterday')
            else:
                text = last_played.strftime('%x')

        return text


class LastPlayedTagFormatter(DateTagFormatter):
    """
    A formatter for the last time a track was played
    """

    def __init__(self):
        DateTagFormatter.__init__(self, '__last_played')


providers.register('tag-formatting', LastPlayedTagFormatter())


class DateAddedTagFormatter(DateTagFormatter):
    """
    A formatter for the date a track
    was added to the collection
    """

    def __init__(self):
        DateTagFormatter.__init__(self, '__date_added')


providers.register('tag-formatting', DateAddedTagFormatter())


class LocationTagFormatter(TagFormatter):
    """
    A formatter for the location of a track,
    properly sanitized if necessary
    """

    def __init__(self):
        TagFormatter.__init__(self, '__loc')

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: optionally passed parameters
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        path = track.get_local_path()
        if path is not None:
            return path
        return common.sanitize_url(track.get_tag_raw('__loc'))


providers.register('tag-formatting', LocationTagFormatter())


class CommentTagFormatter(TagFormatter):
    """
    A formatter for comments embedded in tracks
    """

    def __init__(self):
        TagFormatter.__init__(self, 'comment')

    def format(self, track, parameters):
        """
        Formats a raw tag value

        :param track: the track to get the tag from
        :type track: :class:`xl.trax.Track`
        :param parameters: whether to keep newlines,
            possible values for "newlines" are:

            * keep: do not strip newlines (default)
            * strip: strip newlines
        :type parameters: dictionary
        :returns: the formatted value
        :rtype: string
        """
        value = track.get_tag_raw(self.name)

        if not value:
            return ''

        value = '\n\n'.join(value)
        newlines = parameters.get('newlines', 'keep')

        return self.format_value(value, newlines)

    @staticmethod
    def format_value(value, newlines='keep'):
        """
        Formats a comment value

        :param value: the comment text
        :type value: string
        :param newlines: whether to keep newlines,
            possible values are:

            * keep: do not strip newlines (default)
            * strip: strip newlines

        :type newlines: string
        :returns: the formatted value
        :rtype: string
        """
        if newlines == 'strip':
            value = ' '.join(value.splitlines())

        return value


providers.register('tag-formatting', CommentTagFormatter())

# vim: et sts=4 sw=4



from gi.repository import GLib
from gi.repository import Gtk

import http.client
import logging
import operator
import os
import re
import socket
import urllib.parse
from xml.dom import minidom

from xl import common, event, playlist, xdg
from xl.radio import RadioStation, RadioList, RadioItem
from xl.nls import gettext as _
from xlgui.widgets import dialogs


logger = logging.getLogger(__name__)
STATION = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, "exaile_loaded")
    else:
        _enable(None, exaile, None)


def _enable(devicename, exaile, nothing):
    global STATION

    STATION = IcecastRadioStation(exaile)
    exaile.radio.add_station(STATION)


def disable(exaile):
    global STATION
    exaile.radio.remove_station(STATION)
    STATION = None


def set_status(message, timeout=0):
    from xlgui.panel import radio

    radio.set_status(message, timeout)


class IcecastRadioStation(RadioStation):
    """
    Icecast Radio

    Simple usage:

    >>> manager = RadioManager()
    >>> manager.add_station(IcecastRadioStation())
    >>> stations = manager.search('icecast', 'ravetrax')
    >>> tracks = stations[0].get_playlist().get_tracks()
    >>> len(tracks) > 0
    True
    >>>
    """

    name = 'icecast'

    def __init__(self, exaile):
        """
        Initializes the icecast radio station
        """
        self.exaile = exaile
        self.user_agent = exaile.get_user_agent_string('icecast')
        self.icecast_url = 'http://dir.xiph.org'
        self.genre_url = self.icecast_url + '/by_genre'
        self.search_url_prefix = self.icecast_url + '/search?search='
        self.cache_file = os.path.join(xdg.get_cache_dir(), 'icecast.cache')
        self.data = {}
        self._load_cache()
        self.subs = {}
        self.playlists = {}

        logger.debug(self.user_agent)

    def _load_cache(self):
        """
        Loads icecast data from cache
        """
        self.data = {}
        if os.path.isfile(self.cache_file):
            dom = minidom.parse(self.cache_file)
            for genre in dom.getElementsByTagName('genre'):
                self.data[genre.getAttribute('name')] = genre.getAttribute('location')

    def _save_cache(self):
        """
        Saves cache data
        """
        impl = minidom.getDOMImplementation()
        document = impl.createDocument(None, 'genrelist', None)
        genrelist = document.documentElement
        for k, v in self.data.items():
            genre = document.createElement('genre')
            genre.setAttribute('name', k)
            genre.setAttribute('location', v)
            genrelist.appendChild(genre)
        with open(self.cache_file, 'w') as h:
            document.writexml(h, indent='\n', encoding='utf-8')

    def get_lists(self, no_cache=False):
        """
        Returns the rlists for icecast
        """
        from xlgui.panel import radio

        if no_cache or not self.data:
            set_status(_('Contacting Icecast server...'))
            hostinfo = urllib.parse.urlparse(self.genre_url)
            c = http.client.HTTPConnection(hostinfo.netloc, timeout=20)
            try:
                c.request('GET', hostinfo.path, headers={'User-Agent': self.user_agent})
                response = c.getresponse()
            except (socket.timeout, socket.error):
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            if response.status != 200:
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            body = response.read()
            c.close()
            set_status('')

            data = {}
            dom = minidom.parseString(body)
            divs = dom.getElementsByTagName('div')
            for div in divs:
                if div.getAttribute('id') == 'content':
                    anchors = div.getElementsByTagName('a')
                    for anchor in anchors:
                        anchor.normalize()
                        for node in anchor.childNodes:
                            if node.nodeType == minidom.Node.TEXT_NODE:
                                data[node.nodeValue] = anchor.getAttribute('href')
                                break
                    break
            self.data = data
            self._save_cache()
        else:
            data = self.data
        rlists = []

        for item in data.keys():
            rlist = RadioList(item, station=self)
            rlist.get_items = lambda no_cache, name=item: self._get_subrlists(
                name=name, no_cache=no_cache
            )
            rlists.append(rlist)

        rlists.sort(key=operator.attrgetter('name'))
        self.rlists = rlists
        return rlists

    def _get_subrlists(self, name, no_cache=False):
        """
        Gets the subrlists for a rlist
        """
        if name in self.subs and not no_cache:
            return self.subs[name]

        url = self.icecast_url + self.data[name]

        rlists = self._get_stations(url)
        rlists.sort(key=operator.attrgetter('name'))

        self.subs[name] = rlists
        return rlists

    def _get_playlist(self, name, station_id):
        """
        Gets the playlist for the given name and id
        """
        if station_id in self.playlists:
            return self.playlists[station_id]
        url = self.icecast_url + '/listen/' + station_id + '/listen.xspf'
        set_status(_('Contacting Icecast server...'))

        self.playlists[station_id] = playlist.import_playlist(url)
        set_status('')
        return self.playlists[station_id]

    def search(self, keyword):
        """
        Searches the station for a specified keyword

        @param keyword: the keyword to search
        """
        url = self.search_url_prefix + urllib.parse.quote_plus(keyword)
        return self._get_stations(url)

    def _get_stations(self, url):
        from xlgui.panel import radio

        hostinfo = urllib.parse.urlparse(url)
        query = hostinfo.query
        items = []
        thisPage = -1
        nextPage = 0
        set_status(_('Contacting Icecast server...'))
        c = http.client.HTTPConnection(hostinfo.netloc, timeout=20)
        while thisPage < nextPage:
            thisPage += 1
            try:
                c.request(
                    'GET',
                    "%s?%s" % (hostinfo.path, query),
                    headers={'User-Agent': self.user_agent},
                )
                response = c.getresponse()
            except (socket.timeout, socket.error):
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            if response.status != 200:
                raise radio.RadioException(_('Error connecting to Icecast server.'))

            body = response.read().decode('utf-8', 'replace')

            # XML parser can't handle the audio tag
            body = re.sub('<audio.*?>.*?</audio>', '', body, flags=(re.M | re.DOTALL))

            dom = minidom.parseString(body)
            divs = dom.getElementsByTagName('div')
            for div in divs:
                if div.getAttribute('id') == 'content':
                    servers = div.getElementsByTagName('tr')
                    for server in servers:
                        spans = server.getElementsByTagName('span')
                        for span in spans:
                            if span.getAttribute('class') == 'name':
                                span.normalize()
                                if span.firstChild.nodeType == minidom.Node.TEXT_NODE:
                                    sname = span.firstChild.nodeValue
                                else:
                                    sname = span.firstChild.firstChild.nodeValue
                                break
                        tds = server.getElementsByTagName('td')
                        for td in tds:
                            if td.getAttribute('class') == 'tune-in':
                                anchors = td.getElementsByTagName('a')
                                for anchor in anchors:
                                    href = anchor.getAttribute('href')
                                    matcher = re.match(
                                        '/listen/(\d+)/listen\.xspf\Z', href
                                    )
                                    if matcher:
                                        sid = matcher.group(1)
                                        break
                                paragraphs = td.getElementsByTagName('p')
                                for paragraph in paragraphs:
                                    if paragraph.hasAttribute('title'):
                                        quality = paragraph.getAttribute(
                                            'title'
                                        ).split()
                                        if quality[0] == 'Quality':
                                            sbitrate = self._calc_bitrate(quality[1])
                                        elif len(quality[0]) > 3:
                                            sbitrate = str(int(quality[0]) // 1024)
                                        else:
                                            sbitrate = quality[0]
                                        anchor = paragraph.getElementsByTagName(
                                            'a'
                                        ).item(0)
                                        anchor.normalize()
                                        for text in anchor.childNodes:
                                            if text.nodeType == minidom.Node.TEXT_NODE:
                                                sformat = text.nodeValue
                                                break
                                        break
                                break
                        items.append((sname, sid, sbitrate, sformat))

                    nextPage = -1
                    uls = div.getElementsByTagName('ul')
                    for ul in uls:
                        if ul.getAttribute('class') == 'pager':
                            anchors = ul.getElementsByTagName('a')
                            query = anchors.item(anchors.length - 1).getAttribute(
                                'href'
                            )
                            matcher = re.match('\?(.*?page=(\d+))\Z', query)
                            query = matcher.group(1)
                            nextPage = int(matcher.group(2))
                            break
                    break
            dom.unlink()
        c.close()
        set_status('')
        rlists = []

        for item in items:
            rlist = RadioItem(item[0], station=self)
            rlist.bitrate = item[2]
            rlist.format = item[3]
            rlist.get_playlist = lambda name=item[0], station_id=item[
                1
            ]: self._get_playlist(name, station_id)
            rlists.append(rlist)

        return rlists

    def _calc_bitrate(self, quality):
        q = float(quality.replace(',', '.'))
        if q < 5.0:
            bitrate = 64.0 + q * 16.0
        elif q < 9.0:
            bitrate = 160.0 + (q - 5.0) * 32.0
        else:
            bitrate = 320.0 + (q - 9.0) * 180.0
        return str(int(bitrate))

    def on_search(self):
        """
        Called when the user wants to search for a specific stream
        """
        dialog = dialogs.TextEntryDialog(
            _("Enter the search keywords"), _("Icecast Search")
        )

        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            keyword = dialog.get_value()

            self.do_search(keyword)

    @common.threaded
    def do_search(self, keyword):
        """
        Actually performs the search in a separate thread
        """
        lists = self.search(keyword)

        GLib.idle_add(self.search_done, keyword, lists)

    @common.idle_add()
    def search_done(self, keyword, lists):
        """
        Called when the search is finished
        """
        if not lists:
            dialogs.info(self.exaile.gui.main.window, _('No Stations Found'))
            return

        dialog = ResultsDialog(_("Icecast Search Results"))
        dialog.set_items(lists)
        dialog.connect('response', self._search_response)
        dialog.show_all()
        self._keyword = keyword

    def _search_response(self, dialog, result, *e):

        dialog.hide()
        if result == Gtk.ResponseType.OK:
            items = dialog.get_items()
            if not items:
                return

            self.do_get_playlist(self._keyword, items[0])

    @common.threaded
    def do_get_playlist(self, keyword, item):
        pl = item.get_playlist()
        if not pl:
            return

        GLib.idle_add(self.done_getting_playlist, pl)

    @common.idle_add()
    def done_getting_playlist(self, pl):
        self._parent.emit('playlist-selected', pl)

    def get_menu(self, parent):
        """
        Returns a menu that works for icecast radio
        """
        self._parent = parent
        menu = parent.get_menu()
        menu.add_simple(_("Search"), lambda *e: self.on_search(), Gtk.STOCK_FIND)
        return menu


class ResultsDialog(dialogs.ListDialog):
    def __init__(self, title):
        dialogs.ListDialog.__init__(self, title)
        col = self.list.get_column(0)
        col.set_title(_('Name'))
        col.set_expand(True)
        col.set_resizable(True)
        self.list.set_headers_visible(True)
        text = Gtk.CellRendererText()
        text.set_property('xalign', 1.0)
        col = Gtk.TreeViewColumn(_('Bitrate'), text)
        col.set_cell_data_func(
            text,
            lambda column, cell, model, iter: cell.set_property(
                'text', model.get_value(iter, 0).bitrate
            ),
        )
        self.list.append_column(col)
        text = Gtk.CellRendererText()
        text.set_property('xalign', 0.5)
        col = Gtk.TreeViewColumn(_('Format'), text)
        col.set_cell_data_func(
            text,
            lambda column, cell, model, iter: cell.set_property(
                'text', model.get_value(iter, 0).format
            ),
        )
        self.list.append_column(col)



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

import re

from xl import common, settings
from xl.nls import gettext as _

from xlgui.widgets import dialogs, menu, notebook

from . import gt_common

#
# GroupTaggerView signal 'changed' enum
#

group_change = common.enum(
    added=object(), deleted=object(), edited=object()
)  # edited/toggled group

category_change = common.enum(
    added=object(),
    deleted=object(),
    expanded=object(),  # user expanded category display
    collapsed=object(),  # user collapsed category display
    updated=object(),
)  # added group, changed name

# default group category
uncategorized = _('Uncategorized')


class GTShowTracksMenuItem(menu.MenuItem):
    def __init__(self, name, after):
        menu.MenuItem.__init__(self, name, None, after)

    def factory(self, menu, parent, context):

        groups = context['groups']

        if len(groups) == 0:
            display_name = _('Show tracks with selected')
        elif len(groups) == 1:
            display_name = _('Show tracks tagged with "%s"') % groups[0]
        else:
            display_name = _('Show tracks with all selected')

        menuitem = Gtk.MenuItem.new_with_mnemonic(display_name)
        menuitem.connect(
            'activate',
            lambda *e: gt_common.create_all_search_playlist(
                context['groups'], parent.exaile
            ),
        )
        return menuitem


class GroupTaggerContextMenu(menu.Menu):
    def __init__(self, tagger):
        menu.Menu.__init__(self, tagger)

    def get_context(self):
        return self._parent.get_context()


class GroupTaggerView(Gtk.TreeView):
    '''Treeview widget to display tag lists'''

    __gsignals__ = {
        'category-changed': (
            GObject.SignalFlags.ACTION,
            None,
            (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING),
        ),
        'category-edited': (
            GObject.SignalFlags.ACTION,
            None,
            (GObject.TYPE_STRING, GObject.TYPE_STRING),
        ),
        'group-changed': (
            GObject.SignalFlags.ACTION,
            None,
            (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING),
        ),
        'group-edited': (
            GObject.SignalFlags.ACTION,
            None,
            (GObject.TYPE_STRING, GObject.TYPE_STRING),
        ),
    }

    def __init__(self, exaile, model=None, editable=False):

        super(GroupTaggerView, self).__init__()

        self.exaile = exaile

        self.connect('notify::model', self.on_notify_model)

        self.set_model(model)
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        self._row_expanded_id = self.connect('row-expanded', self.on_row_expanded)
        self._row_collapsed_id = self.connect('row-collapsed', self.on_row_collapsed)

        if editable:
            self.set_reorderable(True)

        # Setup the first column, not shown by default
        cell = Gtk.CellRendererToggle()
        cell.set_property('mode', Gtk.CellRendererMode.ACTIVATABLE)
        cell.set_activatable(True)
        cell.connect('toggled', self.on_toggle)

        self.click_column = Gtk.TreeViewColumn(None, cell, active=0, visible=2)

        # Setup the second column
        cell = Gtk.CellRendererText()
        cell.set_property('editable', editable)
        if editable:
            cell.connect('edited', self.on_edit)

        self.text_column = cell
        self.append_column(
            Gtk.TreeViewColumn(_('Tag'), self.text_column, text=1, weight=3)
        )

        #
        # Menu setup
        #

        self.menu = GroupTaggerContextMenu(self)
        smi = menu.simple_menu_item
        sep = menu.simple_separator

        self.connect('popup-menu', self.on_popup_menu)
        self.connect('button-press-event', self.on_button_press)

        if editable:

            item = smi('addgrp', [], _('Add new tag'), callback=self.on_menu_add_group)
            self.menu.add_item(item)

            item = smi(
                'delgrp',
                ['addgrp'],
                _('Delete tag'),
                callback=self.on_menu_delete_group,
                condition_fn=lambda n, p, c: False if len(c['groups']) == 0 else True,
            )
            self.menu.add_item(item)

            self.menu.add_item(sep('sep1', ['delgrp']))

            item = smi(
                'addcat',
                ['sep1'],
                _('Add new category'),
                callback=self.on_menu_add_category,
            )
            self.menu.add_item(item)

            item = smi(
                'remcat',
                ['addcat'],
                _('Remove category'),
                callback=self.on_menu_del_category,
                condition_fn=lambda n, p, c: False
                if len(c['categories']) == 0
                else True,
            )
            self.menu.add_item(item)

            self.menu.add_item(sep('sep2', ['remcat']))

        self.menu.add_item(GTShowTracksMenuItem('sel', ['sep2']))

        item = smi(
            'selcust',
            ['sel'],
            _('Show tracks with selected (custom)'),
            callback=lambda w, n, p, c: gt_common.create_custom_search_playlist(
                c['groups'], exaile
            ),
            condition_fn=lambda n, p, c: True if len(c['groups']) > 1 else False,
        )
        self.menu.add_item(item)

        # TODO:
        # - Create smart playlist from selected

    def set_model(self, model):
        super(GroupTaggerView, self).set_model(model)
        # this gets reset each time set_model is called... so we override
        self.set_search_column(1)

    def set_font(self, font):
        self.text_column.set_property('font-desc', font)
        model = self.get_model()
        self.set_model(None)
        self.set_model(model)
        self.sync_expanded()
        self.queue_draw()

    def get_context(self):
        '''Returns context parameter required by menus'''
        context = common.LazyDict(self)
        context[
            'selected-rows'
        ] = lambda name, parent: parent.get_selection().get_selected_rows()
        context['groups'] = lambda name, parent: parent.get_selected_groups(
            context['selected-rows']
        )
        context['categories'] = lambda name, parent: parent.get_selected_categories(
            context['selected-rows']
        )
        return context

    def show_click_column(self):
        if len(self.get_columns()) == 1:
            self.insert_column(self.click_column, 0)

    def hide_click_column(self):
        if len(self.get_columns()) == 2:
            self.remove_column(self.click_column)

    def on_notify_model(self, object, property_spec):
        model = self.get_model()
        if model:
            model.connect('row-changed', self.on_row_changed)
            model.connect('row-deleted', self.on_row_deleted)

        # TODO: what's the best way to disconnect when it's unset or changed?

    def on_edit(self, cell, path, new_text):
        if new_text != "":
            model = self.get_model()
            old = model.change_name(path, new_text)

            if model.is_category(path):
                self.emit('category-edited', old, new_text)
            else:
                self.emit('group-changed', group_change.edited, old)

    def on_row_changed(self, model, path, iter):
        if self.get_model() and not model.is_category(path):
            category = model.get_category(path)
            if category is not None:
                self.emit('category-changed', category_change.updated, category)
                self.expand_to_path(path)

    def on_row_collapsed(self, widget, iter, path):
        self.emit(
            'category-changed',
            category_change.collapsed,
            self.get_model().get_category(path),
        )

    def on_row_deleted(self, model, path):
        if self.get_model() and not model.is_category(path):
            category = model.get_category(path)
            if category is not None:
                self.emit('category-changed', category_change.updated, category)

    def on_row_expanded(self, widget, iter, path):
        self.emit(
            'category-changed',
            category_change.expanded,
            self.get_model().get_category(path),
        )

    def on_toggle(self, cell, path):
        self.get_model()[path][0] = not cell.get_active()
        self.emit('group-changed', group_change.edited, None)

    def on_menu_add_group(self, widget, name, parent, context):
        # TODO: instead of dialog, just add a new thing, make it editable?
        input = dialogs.TextEntryDialog(_('New tag value?'), _('Enter new tag value'))

        if input.run() == Gtk.ResponseType.OK:
            group = input.get_value()

            if group != "":
                model, paths = context['selected-rows']

                categories = context['categories']
                if len(categories):
                    category = categories[0]
                else:
                    category = uncategorized

                if model.add_group(group, category, True):
                    self.emit('group-changed', group_change.added, group)
                else:
                    self.emit('group-changed', group_change.edited, None)

                self.emit('category-changed', category_change.updated, category)

    def on_menu_delete_group(self, widget, name, parent, context):
        '''Menu says delete something'''

        model, paths = context['selected-rows']
        groups = model.delete_selected_groups(paths)

        for group in groups:
            self.emit('group-changed', group_change.deleted, group)

    def on_menu_add_category(self, widget, name, parent, context):
        # TODO: instead of dialog, just add a new thing, make it editable?
        input = dialogs.TextEntryDialog(
            _('New Category?'), _('Enter new category name')
        )

        if input.run() == Gtk.ResponseType.OK:
            category = input.get_value()

            if category != "":
                model, paths = context['selected-rows']
                if model.add_category(category):
                    self.emit('category-changed', category_change.added, category)

    def on_menu_del_category(self, widget, name, parent, context):

        model, paths = context['selected-rows']
        categories = model.delete_selected_categories(paths)

        for category, groups in categories.items():
            self.emit('category-changed', category_change.deleted, category)
            for group in groups:
                self.emit('group-changed', group_change.deleted, group)

    def get_selected_groups(self, selected_rows):
        model, rows = selected_rows
        return model.get_selected_groups(rows)

    def get_selected_categories(self, selected_rows):
        model, rows = selected_rows
        return model.get_selected_categories(rows)

    def on_button_press(self, widget, event):
        widget.do_button_press_event(widget, event)
        if event.triggers_context_menu():
            self.menu.popup(None, None, None, None, event.button, event.time)
        return True

    def on_popup_menu(self, widget):
        self.menu.popup(None, None, None, None, 0, 0)
        return True

    def sync_expanded(self):
        '''Syncs the expansion state stored in the model to the tree'''
        with self.handler_block(self._row_expanded_id):
            with self.handler_block(self._row_collapsed_id):
                for row in self.get_model():
                    if row[0]:
                        self.expand_row(row.path, True)


class GroupTaggerTreeStore(Gtk.TreeStore, Gtk.TreeDragSource, Gtk.TreeDragDest):
    """
    The tree model for grouptagger

    Rows for categories:
        [expanded, category name, False]
    Rows for groups:
        [selected, group name, True]
    """

    def __init__(self):
        super(GroupTaggerTreeStore, self).__init__(
            GObject.TYPE_BOOLEAN,
            GObject.TYPE_STRING,
            GObject.TYPE_BOOLEAN,
            GObject.TYPE_INT,
        )
        self.set_sort_column_id(1, Gtk.SortType.ASCENDING)

    def add_category(self, category):
        '''Returns True if added new category, False otherwise'''

        for row in self:
            if row[1] == category:
                return False

        self.append(None, [True, category, False, Pango.Weight.BOLD])
        return True

    def add_group(self, group, category=uncategorized, selected=True):
        '''Returns True if added new group, False otherwise'''

        for row in self:
            if row[1] == category:
                for chrow in row.iterchildren():
                    if chrow[1] == group:
                        row[0] = selected
                        return False

                self.append(row.iter, [selected, group, True, Pango.Weight.NORMAL])
                return True

        # add new category
        it = self.append(None, [True, category, False, Pango.Weight.BOLD])
        # add value to that category
        self.append(it, [selected, group, True, Pango.Weight.NORMAL])
        return True

    def change_name(self, path, name):
        old = self[path][1]
        self[path][1] = name
        return old

    def delete_selected_categories(self, paths):

        categories = {}
        iters = [self.get_iter(path) for path in paths if self[path].parent is None]

        for i in iters:
            if i is not None:
                groups = []
                for ch in self[i].iterchildren():
                    groups.append(ch[1])
                categories[self.get_value(i, 1)] = groups
                self.remove(i)

        return categories

    def delete_selected_groups(self, paths):
        '''Deletes selected groups, returns a list of the removed groups'''
        groups = []
        iters = [self.get_iter(path) for path in paths if self[path].parent is not None]

        for i in iters:
            if i is not None:
                groups.append(self.get_value(i, 1))
                self.remove(i)

        return groups

    def get_category(self, path):
        '''Given a path, return the category associated with that path'''
        if len(path) == 1:
            if len(self):
                return self[path][1]
        else:
            return self[(path[0],)][1]

    def get_category_groups(self, category):
        return [row[1] for row in self.iter_category(category)]

    def get_selected_groups(self, paths):
        '''rows is obtained from get_selection().get_rows()'''
        return [self[path][1] for path in paths if self[path].parent is not None]

    def get_selected_categories(self, paths):
        '''rows is obtained from get_selection().get_rows()'''
        return [self[path][1] for path in paths if self[path].parent is None]

    def is_category(self, path):
        return len(path) == 1

    def iter_active(self):
        '''Iterate over all groups with the checkbox on'''
        for row in self.iter_group_rows():
            if row[0]:
                yield row[1]

    def iter_category(self, category):
        '''Iterate over all rows of a category'''
        for row in self:
            if category == row[1]:
                for chrow in row.iterchildren():
                    yield chrow
                break

    def iter_group_rows(self):
        '''Iterate over all groups in the model, yields (selected, group, other)'''
        for row in self:
            for chrow in row.iterchildren():
                yield chrow

    def iter_groups(self):
        '''Iterate over all groups in the model, yields each group'''
        for row in self.iter_group_rows():
            yield row[1]

    # def has_group(self, group):
    #    for row in self:
    #        for chrow in row:
    #            if row[1] == group:
    #                return True
    #    return False

    def load(self, group_categories):
        """
        input format:

        { category: [expanded, [(active, group), ... ]], ... }
        """
        for category, (expanded, groups) in group_categories.items():
            cat = self.append(None, [expanded, category, False, Pango.Weight.BOLD])
            for active, group in groups:
                self.append(cat, [active, group, True, Pango.Weight.NORMAL])

    #
    # DND interface
    #

    def do_row_draggable(self, path):
        '''Only groups are draggable'''
        return self[path].parent is not None

    def do_row_drop_possible(self, dest_path, selection_data):
        '''Can only drag to different categories'''
        _, _, src_path = Gtk.tree_get_row_drag_data(selection_data)
        return len(dest_path) == 2 and src_path[0] != dest_path[0]


class GroupTaggerWidget(Gtk.Box):
    '''Melds the tag view with an 'add' button'''

    def __init__(self, exaile):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        self.title = Gtk.Label()
        self.artist = Gtk.Label()
        self.view = GroupTaggerView(exaile, GroupTaggerTreeStore(), editable=True)
        self.store = self.view.get_model()

        self.tag_button = Gtk.Button(label=_('Add Tag'))
        self.tag_button.connect('clicked', self.on_add_tag_click)

        self.title.set_xalign(0)
        self.title.set_yalign(0.5)
        self.title.set_line_wrap(True)
        self.title.set_property('wrap-mode', Pango.WrapMode.WORD_CHAR)
        self.title.hide()

        self.artist.set_xalign(0)
        self.artist.set_yalign(0.5)
        self.artist.set_line_wrap(True)
        self.artist.set_property('wrap-mode', Pango.WrapMode.WORD_CHAR)
        self.artist.hide()

        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        scroll.add(self.view)

        self.pack_start(self.title, False, True, 0)
        self.pack_start(self.artist, False, True, 0)
        self.pack_start(scroll, True, True, 0)
        self.pack_start(self.tag_button, False, True, 0)

    def on_add_tag_click(self, widget):
        self.view.on_menu_add_group(self.view, None, None, self.view.get_context())

    def set_font(self, font):
        self.view.set_font(font)

    def set_title(self, title):
        '''Sets the title for this widget. Hides title if title is None'''
        if title is None:
            self.title.hide()
        else:
            self.title.set_markup(
                '<big><b>' + GLib.markup_escape_text(title) + '</b></big>'
            )
            self.title.show()

    def set_artist(self, artist):
        '''Sets the author for this widget. Hides it if author is None'''
        if artist is None:
            self.artist.hide()
        else:
            self.artist.set_text('by ' + artist)
            self.artist.show()

    def set_track_info(self, track):
        '''Shortcut for set_title/set_artist'''
        if track is None:
            self.set_title(None)
            self.set_artist(None)
        else:
            self.set_title(track.get_tag_display('title'))
            self.set_artist(track.get_tag_display('artist'))

    def add_groups(self, groups):

        added = False

        self.view.freeze_child_notify()
        self.view.set_model(None)

        for group in groups:
            added = self.store.add_group(group, uncategorized, selected=False) or added

        self.view.set_model(self.store)
        self.view.sync_expanded()

        if added:
            self.view.emit('category-changed', category_change.updated, uncategorized)

        self.view.thaw_child_notify()

    def set_categories(self, groups, group_categories):
        """
        groups: iterable
        group_categories: dict: key is category, value is (visible, list of groups)
        """

        defaults = {}
        set_groups = set()  # this holds all groups that were found

        # validate it
        for category, (visible, cgroups) in group_categories.items():
            dcgroups = []
            for group in cgroups:
                if group not in set_groups:
                    dcgroups.append((group in groups, group))
                    set_groups.add(group)

            defaults[category] = (visible, dcgroups)

        # Add anything left over to uncategorized
        groups = set(groups).difference(set_groups)
        if len(groups):
            other = defaults.setdefault(uncategorized, (True, []))
            other[1].extend([(True, group) for group in groups])

        self.view.freeze_child_notify()
        self.view.set_model(None)

        self.store.clear()

        self.store.load(defaults)

        self.view.set_model(self.store)
        self.view.sync_expanded()

        self.view.thaw_child_notify()


class GroupTaggerPanel(notebook.NotebookPage):
    '''A panel that has all of the functionality in it'''

    menu_provider_name = 'panel-tab-context'

    def __init__(self, exaile):

        notebook.NotebookPage.__init__(self)

        # add the tagger widget
        self.tagger = GroupTaggerWidget(exaile)

        # add the widgets to this page
        self.pack_start(self.tagger, True, True, 0)

        self.name = 'grouptagger'

    def get_page_name(self):
        return _('GroupTagger')

    def get_panel(self):
        return self


class AllTagsListView(Gtk.TreeView):
    def __init__(self, model=None):
        Gtk.TreeView.__init__(self, model)

        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        # Setup the first column
        cell = Gtk.CellRendererToggle()
        cell.set_property('mode', Gtk.CellRendererMode.ACTIVATABLE)
        cell.set_activatable(True)
        cell.connect('toggled', self.on_toggle)

        self.append_column(Gtk.TreeViewColumn(None, cell, active=0))

        # Setup the second column
        self.append_column(
            Gtk.TreeViewColumn(_('Group'), Gtk.CellRendererText(), text=1)
        )

        self.connect('key_press_event', self.on_key_press)

    def on_key_press(self, widget, event):
        if event.keyval == Gdk.KEY_space:
            model, paths = self.get_selection().get_selected_rows()

            sel = False
            for path in paths:
                sel = model[path][0] or sel

            for path in paths:
                model[path][0] = not sel

    def on_toggle(self, cell, path):
        self.get_model()[path][0] = not cell.get_active()


class AllTagsListStore(Gtk.ListStore):
    def __init__(self):
        Gtk.ListStore.__init__(self, GObject.TYPE_BOOLEAN, GObject.TYPE_STRING)
        self.set_sort_column_id(1, Gtk.SortType.ASCENDING)

    def add_group(self, group):
        self.append((False, group))

    def get_active_groups(self):
        return [row[1] for row in self if row[0]]


class AllTagsDialog(Gtk.Window):
    def __init__(self, exaile, callback):

        Gtk.Window.__init__(self)
        self.set_title(_('Get all tags from collection'))
        self.set_resizable(True)
        self.set_size_request(150, 400)

        self.add(Gtk.Frame())

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self._callback = callback

        self.model = AllTagsListStore()
        self.view = AllTagsListView()

        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        scroll.add(self.view)
        scroll.hide()

        vbox.pack_start(scroll, True, True, 0)

        button = Gtk.Button(_('Add selected to choices'))
        button.connect('clicked', self.on_add_selected_to_choices)
        vbox.pack_end(button, False, False, 0)

        self.get_child().add(vbox)

        # get the collection groups
        groups = gt_common.get_all_collection_groups(exaile.collection)
        for group in groups:
            self.model.add_group(group)

        self.view.set_model(self.model)
        self.show_all()

    def on_add_selected_to_choices(self, widget):
        self._callback(self.model.get_active_groups())


class GroupTaggerQueryDialog(Gtk.Dialog):
    """
    Dialog used to allow the user to select the behavior of the query
    used to filter out tracks that match a particular characteristic
    """

    def __init__(self, groups):

        Gtk.Dialog.__init__(self, title=_('Show tracks with groups'))

        # setup combo box selections
        self.group_model = Gtk.ListStore(GObject.TYPE_STRING)
        groups_set = gt_common.get_groups_from_categories()
        groups_set |= set(groups)

        for group in groups_set:
            self.group_model.append([group])

        self.combo_model = Gtk.ListStore(GObject.TYPE_STRING)
        self.choices = [
            _('Must have this tag [AND]'),
            _('May have this tag [OR]'),
            _('Must not have this tag [NOT]'),
            _('Ignored'),
        ]
        for choice in self.choices:
            self.combo_model.append([choice])

        # setup table
        self.table = Gtk.Grid()

        self.table.attach(Gtk.Label(label=_('Group')), 0, 0, 1, 1)
        self.table.attach(Gtk.Label(label=_('Selected Tracks')), 1, 0, 1, 1)

        # TODO: Scrolled window
        self.combos = []

        # TODO: Add/remove groups to/from table

        for i, group in enumerate(sorted(groups)):

            # label
            gcombo = self._init_combo(self.group_model)
            gcombo.set_active(self._get_group_index(group))
            self.table.attach(gcombo, 0, i + 1, 1, 1)

            # combo
            combo = self._init_combo(self.combo_model)
            combo.set_active(0)
            self.table.attach(combo, 1, i + 1, 1, 1)

            self.combos.append((gcombo, combo))

        self.vbox.pack_start(self.table, True, True, 0)

        self.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK, Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL
        )
        self.show_all()

    def _init_combo(self, model):
        combo = Gtk.ComboBox.new_with_model(model)
        cell = Gtk.CellRendererText()
        combo.pack_start(cell, True)
        combo.add_attribute(cell, 'text', 0)
        return combo

    def _get_group_index(self, group):
        for i, row in enumerate(self.group_model):
            if row[0] == group:
                return i
        return -1

    def get_search_params(self):
        '''Returns (name, search_string) from user selections'''

        and_p = [[], '']  # groups, name, re_string
        or_p = [[], '']
        not_p = [[], '']

        first = True
        tagname = settings.get_option(gt_common.tagname_option, 'grouping')
        name = '%s: ' % (tagname.title())

        # gather the data
        for gcombo, combo in self.combos:

            active_group = self.group_model[gcombo.get_active()][0]
            wsel = self.combo_model[combo.get_active()][0]

            if wsel == self.choices[0]:
                and_p[0].append(active_group)
            elif wsel == self.choices[1]:
                or_p[0].append(active_group)
            elif wsel == self.choices[2]:
                not_p[0].append(active_group)

        # create the AND conditions
        if len(and_p[0]):
            name += ' and '.join(and_p[0])
            first = False

            and_p[1] = ' '.join(
                [
                    '%s~"\\b%s\\b"' % (tagname, re.escape(group.replace(' ', '_')))
                    for group in and_p[0]
                ]
            )

        # create the NOT conditions
        if len(not_p[0]):
            if first:
                name += ' and not '.join(not_p[0])
            else:
                name += ' and ' + ' and '.join(['not ' + p for p in not_p[0]])
            first = False

            not_p[1] = ' ! %s~"%s"' % (
                tagname,
                '|'.join(
                    [
                        '\\b' + re.escape(group.replace(' ', '_')) + '\\b'
                        for group in not_p[0]
                    ]
                ),
            )

        # create the OR conditions
        if len(or_p[0]):
            if first:
                name += ' or '.join(or_p[0])
            elif len(or_p[0]) > 1:
                name += ' and (' + ' or '.join(or_p[0]) + ')'
            else:
                name += ' and ' + ' or '.join(or_p[0])

            or_p[1] = ' %s~"%s"' % (
                tagname,
                '|'.join(
                    [
                        '\\b' + re.escape(group.replace(' ', '_')) + '\\b'
                        for group in or_p[0]
                    ]
                ),
            )

        regex = (and_p[1] + or_p[1] + not_p[1]).strip()

        return (name, regex)


class GroupTaggerAddRemoveDialog(Gtk.Dialog):
    def __init__(self, add, tracks, exaile):

        self.add = add
        self.tracks = tracks

        if self.add:
            Gtk.Dialog.__init__(self, title=_("Add tags to all"))
        else:
            Gtk.Dialog.__init__(self, title=_("Remove tags from all"))

        self.add_buttons(Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY)

        # add the tagger widget
        self.tagger = GroupTaggerWidget(exaile)
        self.tagger.set_artist(None)
        self.tagger.set_title(None)

        # add the widgets to this page
        box = self.get_content_area()
        box.pack_start(self.tagger, True, True, 0)

        self.tagger.view.show_click_column()
        self.show_all()

        # TODO: display the tracks being edited?

    def get_active(self):
        return list(self.tagger.view.get_model().iter_active())



# Copyright (C) 2010 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import Gdk

from xl import event, player, providers
from xl.nls import gettext as _
from xlgui.widgets import playback


class ABRepeatPlugin:
    def __init__(self):
        self.__menu_item = None

    def enable(self, _exaile):
        """
        Enables the plugin
        """
        pass  # needs to be implemented, otherwise xl.plugins will break

    def on_gui_loaded(self):
        self.__menu_item = RepeatSegmentMenuItem()
        providers.register('progressbar-context-menu', self.__menu_item)

    def disable(self, _exaile):
        """
        Disables the plugin
        """
        self.__menu_item.destroy()
        providers.unregister('progressbar-context-menu', self.__menu_item)


plugin_class = ABRepeatPlugin


class RepeatSegmentMenuItem(playback.MoveMarkerMenuItem, providers.ProviderHandler):
    """
    Menu item allowing for insertion of two markers
    to signify beginning and end of the segment to repeat
    """

    def __init__(self):
        playback.MoveMarkerMenuItem.__init__(
            self, 'repeat-segment', [], _('Repeat Segment'), 'media-playlist-repeat'
        )
        providers.ProviderHandler.__init__(self, 'playback-markers')

        self.beginning_marker = playback.Marker()
        self.beginning_marker.name = 'repeat-beginning'
        self.beginning_marker.props.anchor = playback.Anchor.NORTH_WEST
        self.beginning_marker.props.label = _('Repeat Beginning')
        self.end_marker = playback.Marker()
        self.end_marker.name = 'repeat-end'
        self.end_marker.props.anchor = playback.Anchor.NORTH_EAST
        self.end_marker.props.label = _('Repeat End')
        self.end_marker.connect('reached', self.on_end_marker_reached)

        event.add_ui_callback(self.on_playback_track_end, 'playback_track_end')

    def destroy(self):
        """
        Cleanups
        """
        event.remove_callback(self.on_playback_track_end, 'playback_track_end')
        self.clear_markers()

    def factory(self, menu, parent, context):
        """
        Generates the menu item
        """
        item = playback.MoveMarkerMenuItem.factory(self, menu, parent, context)

        markers = (
            providers.get_provider('playback-markers', n)
            for n in ('repeat-beginning', 'repeat-end')
        )

        if player.PLAYER.current is None:
            item.set_sensitive(False)
        elif None not in markers:
            # Disable if the markers have already been set
            item.set_sensitive(False)

        return item

    def clear_markers(self):
        """
        Removes both markers
        """
        for name in ('repeat-beginning', 'repeat-end'):
            marker = providers.get_provider('playback-markers', name)

            if marker is not None:
                providers.unregister('playback-markers', marker)

    def on_activate(self, widget, parent, context):
        """
        Inserts the beginning (A) marker
        """
        self.beginning_marker.props.position = context['current-position']
        providers.register('playback-markers', self.beginning_marker)
        context['current-marker'] = self.beginning_marker

        playback.MoveMarkerMenuItem.on_activate(self, widget, parent, context)

    def on_parent_button_press_event(self, widget, event):
        """
        Finishes or cancels insertion of markers
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            if self.move_finish():
                if providers.get_provider('playback-markers', 'repeat-end') is None:
                    position = event.x / widget.get_allocation().width
                    self.end_marker.props.position = position
                    providers.register('playback-markers', self.end_marker)
                    self.move_begin(self.end_marker)

                return True
        elif event.triggers_context_menu():
            if self.move_cancel():
                self.clear_markers()

                return True

        return False

    def on_end_marker_reached(self, marker):
        """
        Seeks to the beginning marker
        """
        player.PLAYER.set_progress(self.beginning_marker.props.position)

    def on_provider_removed(self, provider):
        """
        Removes the opposite marker if one
        of the two markers is removed
        """
        names = ('repeat-beginning', 'repeat-end')

        if provider.name in names:
            self.clear_markers()

    def on_playback_track_end(self, event_type, player, track):
        """
        Removes both markers
        """
        self.clear_markers()



# Copyright (C) 2011 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xlgui.preferences import widgets


from xl.nls import gettext as _
import os

name = _('History')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'history_preferences.ui')
icon = 'document-open-recent'


# defaults
save_on_exit_default = False
history_length_default = 1000


class SaveOnExitPreference(widgets.CheckPreference):
    default = save_on_exit_default
    name = 'plugin/history/save_on_exit'


class HistoryLengthPreference(widgets.SpinPreference):
    default = history_length_default
    name = 'plugin/history/history_length'



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from xl import event, formatter, main, settings, xdg
from xl.nls import gettext as _
import xlgui
from xlgui import cover, guiutil
from xlgui.widgets import playlist
from xlgui.widgets.playback import PlaybackProgressBar


class TrackInfoPane(Gtk.Bin):
    """
    Displays cover art and track data
    """

    def __init__(self, player):
        Gtk.Bin.__init__(self)
        self.__player = player

        builder = Gtk.Builder()
        builder.add_from_file(xdg.get_data_path('ui', 'widgets', 'track_info.ui'))

        info_box = builder.get_object('info_box')
        info_box.get_parent().remove(info_box)
        self.add(info_box)

        self.__auto_update = False
        self.__display_progress = False
        self.__formatter = formatter.TrackFormatter(
            _(
                '<span size="x-large" weight="bold">$title</span>\n'
                'by $artist\n'
                'from $album'
            )
        )
        self.__formatter.connect('notify::format', self.on_notify_format)
        self.__default_text = '<span size="x-large" ' 'weight="bold">%s</span>\n\n' % _(
            'Not Playing'
        )
        self.__cover_size = None
        self.__timer = None
        self.__track = None

        self.info_label = builder.get_object('info_label')
        self.action_area = builder.get_object('action_area')
        self.progress_box = builder.get_object('progress_box')
        self.playback_image = builder.get_object('playback_image')
        self.progressbar = PlaybackProgressBar(player)
        guiutil.gtk_widget_replace(builder.get_object('progressbar'), self.progressbar)

        self.cover = cover.CoverWidget(builder.get_object('cover_image'))
        self.cover.hide()
        self.cover.set_no_show_all(True)

        self.clear()
        self.__update_widget_state()

    def destroy(self):
        """
        Cleanups
        """
        # Make sure to disconnect callbacks
        self.set_auto_update(False)

        Gtk.Bin.destroy(self)

    def get_auto_update(self):
        """
        Gets whether the info pane shall
        be automatically updated or not

        :rtype: bool
        """
        return self.__auto_update

    def set_auto_update(self, auto_update):
        """
        Sets whether the info pane shall
        be automatically updated or not

        :param auto_update: enable or disable
            automatic updating
        :type auto_update: bool
        """
        if auto_update != self.__auto_update:
            self.__auto_update = auto_update

            p_evts = [
                'playback_player_end',
                'playback_track_start',
                'playback_toggle_pause',
                'playback_error',
            ]
            events = ['track_tags_changed', 'cover_set', 'cover_removed']

            if auto_update:
                for e in p_evts:
                    event.add_ui_callback(getattr(self, 'on_%s' % e), e, self.__player)
                for e in events:
                    event.add_ui_callback(getattr(self, 'on_%s' % e), e)

                self.set_track(self.__player.current)
            else:
                for e in p_evts:
                    event.remove_callback(getattr(self, 'on_%s' % e), e, self.__player)
                for e in events:
                    event.remove_callback(getattr(self, 'on_%s' % e), e)

    def get_cover_size(self):
        """
        Gets the preferred cover size

        :rtype: int
        """
        return self.__cover_size or settings.get_option('gui/cover_width', 100)

    def set_cover_size(self, cover_size):
        """
        Overrides the cover size to display,
        set to None to use global default

        :param cover_size: the preferred cover size
        :type cover_size: int
        """
        self.__cover_size = cover_size

    def get_default_text(self):
        """
        Gets the default text displayed
        when the playback is stopped

        :rtype: string
        """
        return self.__default_text

    def set_default_text(self, default_text):
        """
        Sets the default text displayed
        when the playback is stopped

        :param default_text: the new default text
        :type default_text: string
        """
        self.__default_text = default_text

    def get_display_progress(self):
        """
        Returns whether the progress indicator
        is currently visible or not

        :rtype: bool
        """
        return self.__display_progress

    def set_display_progress(self, display_progress):
        """
        Shows or hides the progress indicator. The
        indicator will not be displayed if the
        currently displayed track is not playing.

        :param display_progress: Whether to show
            or hide the progress indicator
        :type display_progress: bool
        """
        self.__display_progress = display_progress
        self.__update_widget_state()

    def get_info_format(self):
        """
        Gets the current format used
        to display the track data

        :rtype: string
        """
        return self.__formatter.props.format

    def set_info_format(self, format):
        """
        Sets the format used to display the track data

        :param format: the format, see the documentation
            of :class:`string.Template` for details
        :type format: string
        """
        self.__formatter.props.format = format

    def set_track(self, track):
        """
        Updates the data displayed in the info pane

        :param track: A track to take the data from,
            clears the info pane if track is None
        :type track: :class:`xl.trax.Track`
        """
        if track is None:
            self.clear()
            return

        self.__track = track

        self.cover.set_track(track)

        self.info_label.set_markup(self.__formatter.format(track, markup_escape=True))
        self.__update_widget_state()

    def __update_widget_state(self):
        if self.__display_progress:
            if self.__track == self.__player.current and not self.__player.is_stopped():

                if self.__player.is_paused():
                    icon_name = 'media-playback-pause'
                else:
                    icon_name = 'media-playback-start'
                self.playback_image.set_from_icon_name(
                    icon_name, Gtk.IconSize.SMALL_TOOLBAR
                )

            self.progress_box.set_no_show_all(False)
            self.progress_box.set_visible(True)
            self.progressbar.set_visible(True)
        else:
            self.progress_box.set_visible(False)
            self.progress_box.set_no_show_all(True)
            self.progressbar.set_visible(True)

    def clear(self):
        """
        Resets the info pane
        """

        self.cover.set_track(None)
        self.info_label.set_markup(self.__default_text)

        self.__track = None
        self.__update_widget_state()

    def get_action_area(self):
        """
        Retrieves the action area
        at the end of the pane

        :rtype: :class:`Gtk.Box`
        """
        return self.action_area

    def on_notify_format(self, formatter, format):
        """
        Updates the displayed data after format changes
        """
        if self.__track is not None:
            self.set_track(self.__track)

    def on_playback_player_end(self, event, player, track):
        """
        Clears the info pane on playback end
        """
        self.clear()

    def on_playback_track_start(self, event, player, track):
        """
        Updates the info pane on track start
        """
        self.set_track(track)

    def on_playback_toggle_pause(self, event, player, track):
        """
        Updates the info pane on playback pause/resume
        """
        self.set_track(track)

    def on_playback_error(self, event, player, message):
        """
        Clears the info pane on playback errors
        """
        self.clear()

    def on_track_tags_changed(self, event, track, tags):
        """
        Updates the info pane on tag changes
        """
        if (
            self.__player is not None
            and not self.__player.is_stopped()
            and track is self.__track
        ):
            self.set_track(track)

    def on_cover_set(self, event, covers, track):
        """
        Updates the info pane on cover set
        """
        if track is self.__track:
            self.set_track(track)

    def on_cover_removed(self, event, covers, track):
        """
        Updates the info pane on cover removal
        """
        if track is self.__track:
            self.set_track(track)


class ToolTip:
    """
    Custom tooltip class to allow for
    extended tooltip functionality
    """

    def __init__(self, parent, widget):
        """
        :param parent: the parent widget the tooltip
            should be attached to
        :param widget: the tooltip widget to be used
            for the tooltip
        """
        if self.__class__.__name__ == 'ToolTip':
            raise TypeError(
                "cannot create instance of abstract "
                "(non-instantiable) type `ToolTip'"
            )

        self.__widget = widget
        self.__widget.unparent()  # Just to be sure

        parent.set_has_tooltip(True)
        parent.connect('query-tooltip', self.on_query_tooltip)

    def on_query_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        """
        Puts the custom widget into the tooltip
        """
        tooltip.set_custom(self.__widget)

        return True


class TrackToolTip(TrackInfoPane, ToolTip):
    """
    Track specific tooltip class, displays
    track data and progress indicators
    """

    def __init__(self, parent, player):
        """
        :param parent: the parent widget the tooltip
            should be attached to
        """
        TrackInfoPane.__init__(self, player)
        ToolTip.__init__(self, parent, self)

        self.set_border_width(6)
        self.info_label.set_ellipsize(Pango.EllipsizeMode.NONE)
        self.cover.set_no_show_all(False)
        self.cover.show_all()

    def destroy(self):
        """
        Cleanups
        """
        TrackInfoPane.destroy(self)


class StatusbarTextFormatter(formatter.Formatter):
    """
    A text formatter for status indicators
    """

    def __init__(self, format):
        """
        :param format: The initial format, see the documentation
            of string.Template for details
        :type format: string
        """
        formatter.Formatter.__init__(self, format)

        self._substitutions = {
            'collection_count': self.get_collection_count,
            'playlist_count': self.get_playlist_count,
            'playlist_duration': self.get_playlist_duration,
        }

    def get_collection_count(self):
        """
        Retrieves the collection count
        """
        return _('%d in collection') % main.exaile().collection.get_count()

    def get_playlist_count(self, selection='none'):
        """
        Retrieves the count of tracks in either the
        full playlist or the current selection

        :param selection: 'none' for playlist count only,
            'override' for selection count if tracks are selected,
            playlist count otherwise, 'only' for selection count only
        :type selection: string
        """
        page = xlgui.main.get_selected_page()

        if not isinstance(page, playlist.PlaylistPage):
            return ''

        playlist_count = len(page.playlist)
        selection_count = page.view.get_selection_count()

        if selection == 'none':
            count = playlist_count
            text = _('%d showing')
        elif selection == 'override':
            if selection_count > 1:
                count = selection_count
                text = _('%d selected')
            else:
                count = playlist_count
                text = _('%d showing')
        elif selection == 'only':
            if selection_count > 1:
                count = selection_count
                text = _('%d selected')
            else:
                count = 0
        else:
            raise ValueError(
                'Invalid argument "%s" passed to parameter '
                '"selection" for "playlist_count", possible arguments are '
                '"none", "override" and "only"' % selection
            )

        if count == 0:
            return ''

        return text % count

    def get_playlist_duration(self, format='short', selection='none'):
        """
        Retrieves the duration of all tracks in
        the playlist or within the selection

        :param format: Verbosity of the output
            Possible values are short, long or verbose
            yielding "1:02:42", "1h, 2m, 42s" or
            "1 hour, 2 minutes, 42 seconds"
        :type format: string
        :param selection: 'none' for playlist count only,
            'override' for selection count if tracks are selected,
            playlist count otherwise, 'only' for selection count only
        :type selection: string
        """
        page = xlgui.main.get_selected_page()

        if not isinstance(page, playlist.PlaylistPage):
            return ''

        playlist_duration = sum(t.get_tag_raw('__length') or 0 for t in page.playlist)
        selection_tracks = page.view.get_selected_tracks()
        selection_count = len(selection_tracks)
        selection_duration = sum(
            t.get_tag_raw('__length') or 0 for t in selection_tracks
        )

        if selection == 'none':
            duration = playlist_duration
        elif selection == 'override':
            if selection_count > 1:
                duration = selection_duration
            else:
                duration = playlist_duration
        elif selection == 'only':
            if selection_count > 1:
                duration = selection_duration
            else:
                duration = 0
        else:
            raise ValueError(
                'Invalid argument "%s" passed to parameter '
                '"selection" for "playlist_duration", possible arguments are '
                '"none", "override" and "only"' % selection
            )

        if duration == 0:
            return ''

        return formatter.LengthTagFormatter.format_value(duration, format)


class Statusbar:
    """
    Convenient access to multiple status labels
    """

    def __init__(self, status_bar):
        """
        Initialises the status bar
        """
        self.status_bar = status_bar
        self.formatter = StatusbarTextFormatter(
            settings.get_option(
                'gui/statusbar_info_format',
                '${playlist_count:selection=override, suffix= }'
                '${playlist_duration:selection=override, format=long, prefix=(, suffix=)\\, }'
                '$collection_count',
            )
        )

        self.info_label = Gtk.Label()

        box = self.status_bar.get_message_area()

        box.pack_end(self.info_label, False, True, 0)

        self.context_id = self.status_bar.get_context_id('status')
        self.message_ids = []

    def set_status(self, status, timeout=0):
        """
        Sets the status message
        """
        self.message_ids += [self.status_bar.push(self.context_id, status)]

        if timeout > 0:
            GLib.timeout_add_seconds(timeout, self.clear_status)

    def clear_status(self):
        """
        Clears the status message
        """
        try:
            for message_id in self.message_ids:
                self.status_bar.remove_message(self.context_id, message_id)
        except AttributeError:
            for message_id in self.message_ids:
                self.status_bar.remove(self.context_id, message_id)

        del self.message_ids[:]
        self.message_ids = []

    def update_info(self):
        """
        Updates the info label text
        """
        self.info_label.set_label(self.formatter.format())


# TODO: Check if we can get a progress indicator in here somehow
class Splash:
    """
    A splash screen suitable for indicating startup;
    will automatically be destroyed after GUI startup
    """

    def __init__(self):
        builder = Gtk.Builder()
        builder.add_from_file(xdg.get_data_path('ui', 'splash.ui'))

        image = builder.get_object('splash_image')
        image.set_from_file(xdg.get_data_path('images', 'splash.png'))

        self.window = builder.get_object('SplashScreen')

    def destroy(self):
        """
        Destroys the splash screen
        """
        self.window.destroy()

    def show(self):
        """
        Shows the splash screen
        """
        # Show the splash screen without causing startup notification.
        Gtk.Window.set_auto_startup_notification(False)
        self.window.show_all()
        Gtk.Window.set_auto_startup_notification(True)

        # Ensure the splash is completely drawn before moving on
        while Gtk.events_pending():
            Gtk.main_iteration()

    def hide(self):
        """
        Hides the splash screen
        """
        self.window.hide()



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from xl import providers, collection
from xl.hal import Handler
from xl.devices import Device
import dbus
import logging
import os

logger = logging.getLogger(__name__)

PROVIDER = None


def enable(exaile):
    global PROVIDER
    PROVIDER = MassStorageHandler()
    providers.register("hal", PROVIDER)


def disable(exaile):
    global PROVIDER
    providers.unregister("hal", PROVIDER)
    PROVIDER = None


class MassStorageDevice(Device):
    def __init__(self, mountpoints, name=""):
        if len(mountpoints) == 0:
            raise ValueError("Must specify at least one mount point")
        if not name:
            name = mountpoints[0].split(os.sep)[-1]
        Device.__init__(self, name)
        self._mountpoints = mountpoints
        self.mountpoints = []

    def connect(self):
        self.mountpoints = [
            str(x) for x in self._mountpoints if str(x) != "" and os.path.exists(str(x))
        ]
        if self.mountpoints == []:
            raise IOError("Device is not mounted.")
        for mountpoint in self.mountpoints:
            library = self.library_class(mountpoint)
            self.collection.add_library(library)
        self.transfer = collection.TransferQueue(self.collection.get_libraries()[0])
        self.connected = True  # set this here so the UI can react

    def disconnect(self):
        self.collection = collection.Collection(name=self.name)
        self.mountpoints = []
        self.transfer = None
        self.connected = False


class HalMountpoint:
    """
    Class to represent a mountpoint so we can delay HAL
    mountpoint resolution.
    """

    def __init__(self, hal, udi):
        self.hal = hal
        self.udi = udi

    def __str__(self):
        udis = self.hal.hal.FindDeviceStringMatch("info.parent", self.udi)
        for u in udis:
            obj = self.hal.bus.get_object("org.freedesktop.Hal", u)
            dev = dbus.Interface(obj, "org.freedesktop.Hal.Device")
            if dev.GetProperty("volume.is_mounted") is True:
                return str(dev.GetProperty("volume.mount_point"))
        return ""


class MassStorageHandler(Handler):
    name = "massstorage"

    def is_type(self, device, capabilities):
        if "portable_audio_player" in capabilities:
            try:
                if "storage" in device.GetProperty(
                    "portable_audio_player.access_method.protocols"
                ):
                    return 10
            except dbus.exceptions.DBusException as e:
                if not e.get_dbus_name() == "org.freedesktop.Hal.NoSuchProperty":
                    logger.exception("Portable audio player without storage property")
        return 0

    def get_udis(self, hal):
        udis = hal.hal.FindDeviceByCapability("portable_audio_player")
        ret = []
        for udi in udis:
            dev_obj = hal.bus.get_object("org.freedesktop.Hal", udi)
            device = dbus.Interface(dev_obj, "org.freedesktop.Hal.Device")
            if device.PropertyExists(
                "portable_audio_player.access_method.protocols"
            ) and "storage" in device.GetProperty(
                "portable_audio_player.access_method.protocols"
            ):
                ret.append(udi)
        return ret

    def device_from_udi(self, hal, udi):
        mass_obj = hal.bus.get_object("org.freedesktop.Hal", udi)
        mass = dbus.Interface(mass_obj, "org.freedesktop.Hal.Device")
        if "storage" not in mass.GetProperty(
            "portable_audio_player.access_method.protocols"
        ):
            return

        u = mass.GetProperty("portable_audio_player.storage_device")
        mountpoints = [HalMountpoint(hal, u)]

        name = mass.GetProperty("info.vendor") + " " + mass.GetProperty("info.product")
        massdev = MassStorageDevice(mountpoints, name)

        return massdev



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import subprocess
import logging
import os
import shutil
from gi.repository import Gtk

from xl import event, player, settings
from xl.nls import gettext as _
from xlgui.widgets import dialogs

from . import srprefs

logger = logging.getLogger(__name__)

STREAMRIPPER = None


def get_preferences_pane():
    return srprefs


class Streamripper:
    def __init__(self):
        self.savedir = None

    def toggle_record(self, add_call):
        current_track = player.PLAYER.current
        if not current_track:
            return True
        if current_track.is_local():
            logger.warning('Streamripper can only record streams')
            return True

        self.savedir = settings.get_option(
            'plugin/streamripper/save_location', os.getenv('HOME')
        )
        options = []
        options.append('streamripper')
        options.append(player.PLAYER._pipe.get_property('uri'))
        options.append('-D')
        options.append('%A/%a/%T')
        if settings.get_option('plugin/streamripper/single_file', False):
            options.append("-a")
            options.append("-A")
        options.append("-r")
        options.append(settings.get_option('plugin/streamripper/relay_port', '8888'))
        options.append("-d")
        options.append(self.savedir)

        try:
            self.process = subprocess.Popen(
                options, 0, None, subprocess.PIPE, subprocess.PIPE, subprocess.PIPE
            )
        except OSError:
            logger.error('There was an error executing streamripper')
            dialogs.error(
                self.exaile.gui.main.window, _('Error ' 'executing streamripper')
            )
            return True

        if add_call:
            event.add_callback(self.quit_application, 'quit_application')
            event.add_ui_callback(
                self.start_track, 'playback_track_start', player.PLAYER
            )
            event.add_ui_callback(
                self.stop_playback, 'playback_player_end', player.PLAYER
            )
        return False

    def stop_ripping(self):
        try:
            self.process.terminate()
        except Exception:
            pass
        if settings.get_option('plugin/streamripper/delete_incomplete', True):
            try:
                shutil.rmtree(self.savedir + "/incomplete")
            except OSError:
                pass

    def quit_application(self, type, player, track):
        self.stop_ripping()

    def stop_playback(self, type, player, track):
        self.stop_ripping()
        self.button.set_active(False)
        self.remove_callbacks()

    def start_track(self, type, player, track):
        self.stop_ripping()
        if self.toggle_record(False):
            self.button.set_active(False)
            self.remove_callbacks()

    def remove_callbacks(self):
        event.remove_callback(self.quit_application, 'quit_application')
        event.remove_callback(self.start_track, 'playback_track_start', player.PLAYER)
        event.remove_callback(self.stop_playback, 'playback_player_end', player.PLAYER)


class Button(Streamripper):
    def __init__(self, exaile):
        self.exaile = exaile
        self.button = Gtk.ToggleButton()
        self.button.set_relief(Gtk.ReliefStyle.NONE)
        image = Gtk.Image.new_from_icon_name('media-record', Gtk.IconSize.MENU)
        self.button.set_image(image)

        toolbar = self.exaile.gui.play_toolbar
        toolbar.pack_start(self.button, False, False, 0)
        toolbar.reorder_child(self.button, 3)

        self.button.show()

        self.button.connect('button-release-event', self.toggle_button_callback)

    def toggle_button_callback(self, widget, data):
        if widget.get_active():
            self.stop_ripping()
            self.remove_callbacks()
        else:
            if self.toggle_record(True):  # couldn't record stream
                self.button.set_active(True)
                self.remove_callbacks()

    def remove_button(self):
        self.exaile.gui.play_toolbar.remove(self.button)
        self.button.hide()
        self.button.destroy()
        self.remove_callbacks()


def enable(exaile):
    try:  # just test if streamripper is installed
        subprocess.call(['streamripper'], stdout=-1, stderr=-1)
    except OSError:
        raise NotImplementedError('Streamripper is not available.')
        return False

    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, nothing):
    global STREAMRIPPER
    STREAMRIPPER = Button(exaile)


def disable(exaile):
    global STREAMRIPPER
    STREAMRIPPER.remove_button()
    STREAMRIPPER = None



#
# Allows GTK 3 python applications to exit when CTRL-C is raised
# From https://bugzilla.gnome.org/show_bug.cgi?id=622084
#
# Author: Simon Feltman
# License: Presume same as pygobject
#


import sys
import signal
from typing import ClassVar, List

from gi.repository import GLib


class InterruptibleLoopContext:
    """
    Context Manager for GLib/Gtk based loops.

    Usage of this context manager will install a single GLib unix signal handler
    and allow for multiple context managers to be nested using this single handler.
    """

    #: Global stack context loops. This is added to per InterruptibleLoopContext
    #: instance and allows for context nesting using the same GLib signal handler.
    _loop_contexts: ClassVar[List['InterruptibleLoopContext']] = []

    #: Single source id for the unix signal handler.
    _signal_source_id = None

    @classmethod
    def _glib_sigint_handler(cls, user_data):
        context = cls._loop_contexts[-1]
        context._quit_by_sigint = True
        context._loop_exit_func()

        # keep the handler around until we explicitly remove it
        return True

    def __init__(self, loop_exit_func):
        self._loop_exit_func = loop_exit_func
        self._quit_by_sigint = False

    def __enter__(self):
        # Only use unix_signal_add if this is not win32 and there has
        # not already been one.
        if sys.platform != 'win32' and not InterruptibleLoopContext._loop_contexts:
            # Add a glib signal handler
            source_id = GLib.unix_signal_add(
                GLib.PRIORITY_DEFAULT, signal.SIGINT, self._glib_sigint_handler, None
            )
            InterruptibleLoopContext._signal_source_id = source_id

        InterruptibleLoopContext._loop_contexts.append(self)

    def __exit__(self, exc_type, exc_value, traceback):
        context = InterruptibleLoopContext._loop_contexts.pop()
        assert self == context

        # if the context stack is empty and we have a GLib signal source,
        # remove the source from GLib and clear out the variable.
        if (
            not InterruptibleLoopContext._loop_contexts
            and InterruptibleLoopContext._signal_source_id is not None
        ):
            GLib.source_remove(InterruptibleLoopContext._signal_source_id)
            InterruptibleLoopContext._signal_source_id = None

        if self._quit_by_sigint:
            # caught by _glib_sigint_handler()
            raise KeyboardInterrupt



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

# script to turn a plugin dir into a .exz file suitable for distribution

# takes one commandline parameter: the name of the plugin to build, which must
# be a subdirectory of the current directory

# outputs the built plugin to the current directory, overwriting any current
# build of that plugin

import os
import tarfile
from optparse import OptionParser


p = OptionParser()
p.add_option(
    "-c",
    "--compression",
    dest="compression",
    action="store",
    choices=("", "gz", "bz2"),
    default="bz2",
)
p.add_option(
    "-e",
    "--ignore-extension",
    dest="extensions",
    action="append",
    default=(".pyc", ".pyo"),
)
p.add_option("-f", "--ignore-file", dest="files", action="append", default=("test.py"))
p.add_option("-O", "--output", dest="output", action="store", default="")
options, args = p.parse_args()

# allowed values: "", "gz", "bz2"
COMPRESSION = options.compression

# don't add files with these extensions to the archive
IGNORED_EXTENSIONS = options.extensions

# don't add files with this exact name to the archive
IGNORED_FILES = options.files

_ = lambda x: x


for dir in args:

    if not os.path.exists(dir):
        print("No such folder %s" % dir)
        break

    print("Making plugin %s..." % dir)

    if not os.path.exists(os.path.join(dir, "PLUGININFO")):
        print("ERROR: no valid info for %s, skipping..." % dir)
        continue

    f = open(os.path.join(dir, "PLUGININFO"))
    info = {}
    for line in f:
        try:
            key, val = line.split("=", 1)
        except ValueError:
            continue
        key = key.strip()
        val = eval(val)
        info[key] = val
    f.close()

    if "Version" not in info:
        print("ERROR: couldn't get version for %s, skipping..." % dir)
        continue

    tfile = tarfile.open(
        options.output + dir + "-%s.exz" % info["Version"],
        "w:%s" % COMPRESSION,
        format=tarfile.USTAR_FORMAT,
    )

    for fold, subdirs, files in os.walk(dir):
        for file in files:
            stop = False
            for ext in IGNORED_EXTENSIONS:
                if file.endswith(ext):
                    stop = True
                    break
            if stop:
                continue
            for name in IGNORED_FILES:
                if file == name:
                    stop = True
                    break
            if stop:
                continue

            path = os.path.join(fold, file)
            tfile.add(path)

    tfile.close()

print("Done.")



"""
    This plugin will make a dbus call to prevent the computer from suspending
    during music playback

    It uses the gnome SessionManager interface via dbus

    TODO: use Gtk.Application.inhibit() for less error prone inhibition.
"""

try:
    import _thread
except ImportError:
    import _dummy_thread as _thread
import logging
import os

import dbus

from xl.player import adapters, PLAYER

SUSPEND_PLUGIN = None
logger = logging.getLogger(__name__)


def enable(exaile):
    """
    Called when the plugin is enabled
    """
    global SUSPEND_PLUGIN

    if SUSPEND_PLUGIN is None:
        try:
            SUSPEND_PLUGIN = SuspendInhibit()
        except EnvironmentError:
            logger.error('Failed to Acquire Suspend Bus')
            raise
        except NotImplementedError:
            logger.error('Desktop Session not implemented')
            raise

        # allow plugin to finished enabling so that if user returns
        # to gnome plugin will not have to be re-enabled

    logger.info('Suspend Inhibitor Enabled')


def disable(exaile):
    """
    Called when the plugin is disabled
    """
    global SUSPEND_PLUGIN

    if SUSPEND_PLUGIN is not None:
        SUSPEND_PLUGIN.destroy()
        SUSPEND_PLUGIN = None

    logger.info('Suspend Inhibitor Disabled')


class SuspendInhibit:
    """
    Attempt to detect desktop session and initialize appropriate adapter
    """

    def __init__(self):
        # Attempt to detect Desktop Session Type
        session = os.getenv('DESKTOP_SESSION', '').lower()
        # see https://askubuntu.com/questions/72549/how-to-determine-which-window-manager-is-running
        xdg_session = os.getenv('XDG_CURRENT_DESKTOP', '').lower()

        # Attempt to find an adaptor that works
        if 'gnome' in session or 'gnome' in xdg_session:
            self.adapter = GnomeAdapter()
        elif 'kde' in session or 'kde' in xdg_session:
            try:
                self.adapter = PowerManagerAdapter()
            except EnvironmentError:
                # Fall back to powerdevil
                self.adapter = KdeAdapter()
        elif 'xfce' in session or 'xfce' in xdg_session:
            self.adapter = XfceAdapter()
        # TODO implement for LXDE, X-Cinnamon, Unity; systemd-inhibit
        elif session == '' and xdg_session == '':
            logger.warning(
                'Could not detect Desktop Session, will try default \
                    Power Manager then Gnome'
            )
            try:
                self.adapter = PowerManagerAdapter()
            except EnvironmentError:
                # Fall back to Gnome power manager
                self.adapter = GnomeAdapter()
        else:
            raise NotImplementedError(xdg_session)

    def destroy(self):
        self.adapter.destroy()


class SuspendAdapter(adapters.PlaybackAdapter):
    """
    Base class for Desktop Session suspend inhibitors

    Subclasses will have to override the DBus call methods

    Thread safe
    """

    PROGRAM = 'exaile'
    ACTIVITY = 'playing-music'

    def __init__(self, bus_name, object_path, interface):
        try:
            bus = dbus.SessionBus()
            obj = bus.get_object(bus_name, object_path)
            self.iface = dbus.Interface(obj, interface)
            logger.info('Suspend Bus Acquired')
        except dbus.DBusException:
            raise EnvironmentError(bus_name + ' bus not available')

        self.inhibited = False
        self.lock = _thread.allocate_lock()

        # Initialize parent object
        adapters.PlaybackAdapter.__init__(self, PLAYER)

        # Inhibit if player currently playing
        if PLAYER.is_playing():
            self.inhibit()

    def inhibit(self):
        """
        Inhibit user session suspension.

        Make DBus call to inhibit session suspension if
        session suspension not already inhibited.

        If suspending already inhibited call does nothing.

        DBus call returns unique cookie used to later uninhibit.
        """

        with self.lock:
            if not self.inhibited:
                self._dbus_inhibit_call()
                self.inhibited = True
                logger.info('Inhibited Suspend')

    def uninhibit(self):
        """
        Uninhibit user session suspension.

        Make DBus call to uninhibit session suspension if
        session suspension not already uninhibited.

        If suspending already uninhibited call does nothing.

        DBus call Requires cookie returned from original inhibit.
        """

        with self.lock:
            if self.inhibited:
                if self.cookie is not None:
                    self._dbus_uninhibit_call()
                    self.cookie = None
                    self.inhibited = False
                    logger.info('Unihibited Suspend')
                else:
                    logger.error('Cannot Unihibit Suspend without cookie')

    def is_inhibited(self):
        """ Inhibit Status """
        return self.inhibited

    def destroy(self):
        """ Cleanup """
        # Make sure to uninhibit when exiting
        self.uninhibit()
        adapters.PlaybackAdapter.destroy(self)
        logger.debug('Adapter Destroyed')

    """
        Playback Adapter Callbacks
    """

    def on_playback_track_start(self, event, player, track):
        self.inhibit()

    def on_playback_player_end(self, event, player, track):
        self.uninhibit()

    def on_playback_toggle_pause(self, event, player, track):
        if player.is_playing():
            self.inhibit()
        else:
            self.uninhibit()

    def _dbus_inhibit_call(self):
        """
        Override, overriding method must set self.inhibited value
        Make DBus call to inhibit suspend
        Must not block
        """
        raise NotImplementedError('Method not Overridden')

    def _dbus_uninhibit_call(self):
        """
        Override, overriding method must set self.inhibited value
        Make DBus call to uninhibit suspend
        Must not block
        """
        raise NotImplementedError('Method not Overridden')


class PowerManagerAdapter(SuspendAdapter):
    """
    Default Adapter, implemented by most desktop sessions
    Adapter for org.freedesktop.PowerManagement.Inhibit Interface
    Some desktop sesssions use different bus names for this interface
    and have other small variances
    """

    def __init__(
        self,
        bus_name='org.freedesktop.PowerManagement',
        object_name='/org/freedesktop/PowerManagement/Inhibit',
        interface_name='org.freedesktop.PowerManagement.Inhibit',
    ):
        SuspendAdapter.__init__(self, bus_name, object_name, interface_name)

    def _dbus_inhibit_call(self):
        self.cookie = self.iface.Inhibit(self.PROGRAM, self.ACTIVITY)

    def _dbus_uninhibit_call(self):
        self.iface.UnInhibit(self.cookie)


class GnomeAdapter(SuspendAdapter):
    """
    Gnome uses a similar interface to org.freedesktop.PowerManagement
    but is has different bus name, object path and interface name
    The inhibit and uninhibit method signatures are also different
    """

    SUSPEND_FLAG = 8

    def __init__(self):
        SuspendAdapter.__init__(
            self,
            'org.gnome.SessionManager',
            '/org/gnome/SessionManager',
            'org.gnome.SessionManager',
        )

    def _dbus_inhibit_call(self):
        """
        Gnome Interface has more paramters
        """
        self.cookie = self.iface.Inhibit(
            self.PROGRAM, 1, self.ACTIVITY, self.SUSPEND_FLAG
        )

    def _dbus_uninhibit_call(self):
        """
        Gnome Interface has different case
        """
        self.iface.Uninhibit(self.cookie)


class KdeAdapter(PowerManagerAdapter):
    """
    Adapter for when org.freedesktop.PowerManager interface
    located at bus org.kde.powerdevil
    """

    def __init__(self):
        try:
            PowerManagerAdapter.__init__(self)
        except EnvironmentError:
            # Fall back to other bus name
            PowerManagerAdapter.__init__(self, bus_name='org.kde.powerdevil')


class XfceAdapter(PowerManagerAdapter):
    """
    Adapter for org.freedesktop.PowerManagement interface at bus name
    org.xfce.PowerManager
    """

    def __init__(self):
        try:
            PowerManagerAdapter.__init__(self)
        except EnvironmentError:
            # Fall back to other bus name
            PowerManagerAdapter.__init(self, bus_name='org.xfce.PowerManager')



#!/usr/bin/env python3
#
# Copyright (C) 2016 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

#
# This requires click to be installed, which is not an Exaile dependency
#


import copy
import datetime
from dbm import whichdb
import json
import os.path
import pickle
import pprint
import shelve

import bsddb3 as bsddb
import click


class Utf8Unpickler(pickle.Unpickler):
    def __init__(self, *args, **kwargs):
        kwargs['encoding'] = 'utf-8'
        super().__init__(*args, **kwargs)


# For compatibility with Python 2 shelves
shelve.Unpickler = Utf8Unpickler

exaile_db = os.path.join(os.path.expanduser('~'), '.local', 'share', 'exaile',
                         'music.db')
exaile_pickle_protocol = 2


def tracks(data):
    for k, v in data.items():
        if not k.startswith('tracks-'):
            continue
        yield k, v


@click.group()
@click.option('--db', default=exaile_db)
@click.pass_context
def cli(ctx, db):
    '''
        Tool that allows low-level exploration of an Exaile music database
    '''
    # simpler version of trackdb.py
    try:
        d = bsddb.hashopen(db, 'r')
        contents = shelve.Shelf(d, protocol=exaile_pickle_protocol)
    except Exception:
        try:
            contents = shelve.open(db, flag='r', protocol=exaile_pickle_protocol)
        except Exception:
            if os.path.exists(db):
                raise
            else:
                raise click.ClickException("%s does not exist" % db)

    ctx.obj = contents

    def _on_close():
        ctx.obj.close()

    ctx.call_on_close(_on_close)


@cli.command()
@click.pass_obj
@click.pass_context
@click.argument('dbtype')
def cvtdb(ctx, data, dbtype):
    '''
        Only used for testing purposes
    '''

    db = ctx.parent.params['db']
    newdb = db + '.new'

    if dbtype == 'gdbm':
        import dbm.gnu
        new_d = dbm.gnu.open(newdb, 'n')
    elif dbtype == 'dbm':
        import dbm.ndbm
        new_d = dbm.ndbm.open(newdb, 'n')
    elif dbtype == 'dbhash':
        import dbm.bsd
        new_d = dbm.bsd.open(newdb, 'n')
    elif dbtype == 'bsddb':
        new_d = bsddb.hashopen(newdb, 'n')
    elif dbtype == 'dumbdbm':
        import dbm.dumb
        new_d = dbm.dumb.open(newdb, 'n')
    else:
        raise click.ClickException("Invalid type %s" % dbtype)

    new_data = shelve.Shelf(new_d, protocol=exaile_pickle_protocol)

    for k, v in data.items():
        new_data[k] = v

    new_data.sync()
    new_data.close()


@cli.command()
@click.pass_obj
@click.argument('output')
def tojson(data, output):
    '''
        Export Exaile's database to JSON
    '''

    # not really a db type, but useful?
    d = {}
    for k, v in data.items():
        d[k] = v
    with open(output, 'w') as fp:
        json.dump(d, fp, sort_keys=True, indent=4, separators=(',', ': '))


@cli.command()
@click.pass_obj
@click.pass_context
def info(ctx, data):
    '''
        Display summary information about the DB
    '''
    print('DB Type:', whichdb(ctx.parent.params['db']))
    print('Version:', data.get('_dbversion'))
    print('Name   :', data.get('name'))
    print('Key    :', data.get('_key'))
    print("Count  :", len(data))
    print()
    print('Location(s):')
    pprint.pprint(data.get('_serial_libraries'))


@cli.command()
@click.argument('search')
@click.option('-t', '--tag', default='title')
@click.pass_obj
def search(data, tag, search):
    '''
        Search for tracks via the contents of particular tags
    '''
    tag = str(tag)
    for k, tr in tracks(data):
        tr = tr[0]
        val = tr.get(tag)
        if val:
            if isinstance(val, list):
                val = val[0]
            if search in val:
                print('%10s: %s' % (k, val))


def _date_fix(tags):
    tags = copy.deepcopy(tags)
    for t in ['__date_added', '__last_played', '__modified']:
        if t in tags:
            try:
                dt = datetime.datetime.fromtimestamp(tags[t])
                tags[t] = dt.strftime("%Y-%m-%d %H:%M:%S")
            except:
                pass

    return tags


def _print_tr(tr, raw):
    if tr is not None:
        tags, key, attrs = tr
        print("Track key", key)
        print("Tags:")
        if not raw:
            tags = _date_fix(tags)
        pprint.pprint(tags)
        print("Attrs:")
        pprint.pprint(attrs)


@cli.command('track-by-idx')
@click.argument('idx')
@click.option('--raw/--no-raw', default=False)
@click.pass_obj
def track_by_idx(data, idx, raw):
    '''
        pprint a track by its index in the database
    '''
    key = str('tracks-%s' % idx)
    tr = data.get(key)
    _print_tr(tr, raw)


@cli.command()
@click.pass_obj
def tags(data):
    '''
        Display summary information about tags present
    '''

    tags = set()

    for k, v in tracks(data):
        # print(v[0].keys())
        tags.update(set(v[0].keys()))

    print("All tags:")
    pprint.pprint(tags)


if __name__ == '__main__':
    cli()



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import dbus
from fcntl import ioctl
import logging
import os
import struct

from xl.nls import gettext as _
from xl import providers, event
from xl.hal import Handler, UDisksProvider
from xl.devices import Device, KeyedDevice

from xl import playlist, trax, common
import os.path

from . import cdprefs

try:
    import DiscID
    import CDDB

    CDDB_AVAIL = True
except Exception:
    CDDB_AVAIL = False

logger = logging.getLogger(__name__)


TOC_HEADER_FMT = 'BB'
TOC_ENTRY_FMT = 'BBBix'
ADDR_FMT = 'BBB' + 'x' * (struct.calcsize('i') - 3)
CDROMREADTOCHDR = 0x5305
CDROMREADTOCENTRY = 0x5306
CDROM_LEADOUT = 0xAA
CDROM_MSF = 0x02
CDROM_DATA_TRACK = 0x04


class CdPlugin:
    def enable(self, exaile):
        self.__exaile = exaile
        self.__udisks2 = None

    def on_exaile_loaded(self):
        # verify that hal/whatever is loaded, load correct provider
        if self.__exaile.udisks2 is not None:
            self.__udisks2 = UDisks2CdProvider()
            providers.register('udisks2', self.__udisks2)

    def teardown(self, exaile):
        if self.__udisks2 is not None:
            providers.unregister('udisks2', self.__udisks2)
        self.__udisks2 = None
        self.__exaile = None

    def disable(self, exaile):
        self.teardown(exaile)

    def get_preferences_pane(self):
        return cdprefs


plugin_class = CdPlugin


class _CDTrack:
    """
    @ivar track: Track number. Starts with 1, which is used for the TOC and contains data.
    @ivar data: `True` if this "track" contains data, `False` if it is audio
    @ivar minutes: Minutes from begin of CD
    @ivar seconds: Seconds after `minutes`, from begin of CD
    @ivar frames: Frames after `seconds`, from begin of CD
    """

    def __init__(self, entry):
        self.track, adrctrl, _format, addr = struct.unpack(TOC_ENTRY_FMT, entry)
        self.minutes, self.seconds, self.frames = struct.unpack(
            ADDR_FMT, struct.pack('i', addr)
        )

        # adr = adrctrl & 0xf
        ctrl = (adrctrl & 0xF0) >> 4

        self.data = False
        if ctrl & CDROM_DATA_TRACK:
            self.data = True

    def get_frame_count(self):
        return (self.minutes * 60 + self.seconds) * 75 + self.frames


class CDTocParser:
    # based on code from http://carey.geek.nz/code/python-cdrom/cdtoc.py

    def __init__(self, device):
        self.__raw_tracks = []
        self.__read_toc(device)

    def __read_toc(self, device):
        fd = os.open(device, os.O_RDONLY)
        try:
            toc_header = struct.pack(TOC_HEADER_FMT, 0, 0)
            toc_header = ioctl(fd, CDROMREADTOCHDR, toc_header)
            start, end = struct.unpack(TOC_HEADER_FMT, toc_header)

            # All tracks plus leadout
            for trnum in list(range(start, end + 1)) + [CDROM_LEADOUT]:
                entry = struct.pack(TOC_ENTRY_FMT, trnum, 0, CDROM_MSF, 0)
                entry = ioctl(fd, CDROMREADTOCENTRY, entry)
                self.__raw_tracks.append(_CDTrack(entry))
        finally:
            os.close(fd)

    def _get_track_lengths(self):
        """ returns track length in seconds """
        track = self.__raw_tracks[0]
        offset = track.get_frame_count()
        lengths = []
        for track in self.__raw_tracks[1:]:
            frame_end = track.get_frame_count()
            lengths.append((frame_end - offset) // 75)
            offset = frame_end
        return lengths


class CDPlaylist(playlist.Playlist):
    def __init__(self, name=_("Audio Disc"), device=None):
        playlist.Playlist.__init__(self, name=name)

        if not device:
            self.__device = "/dev/cdrom"
        else:
            self.__device = device

        self.open_disc()

    def open_disc(self):

        toc = CDTocParser(self.__device)
        lengths = toc._get_track_lengths()

        songs = []

        for count, length in enumerate(lengths):
            count += 1
            song = trax.Track("cdda://%d/#%s" % (count, self.__device))
            song.set_tags(
                title="Track %d" % count, tracknumber=str(count), __length=length
            )
            songs.append(song)

        self.extend(songs)

        if CDDB_AVAIL:
            self.get_cddb_info()

    @common.threaded
    def get_cddb_info(self):
        try:
            disc = DiscID.open(self.__device)
            self.__device = DiscID.disc_id(disc)
            status, info = CDDB.query(self.__device)
        except IOError:
            return

        if status in (210, 211):
            info = info[0]
            status = 200
        if status != 200:
            return

        (status, info) = CDDB.read(info['category'], info['disc_id'])

        title = info['DTITLE'].split(" / ")
        for i in range(self.__device[1]):
            tr = self[i]
            tr.set_tags(
                title=info['TTITLE' + str(i)].decode('iso-8859-15', 'replace'),
                album=title[1].decode('iso-8859-15', 'replace'),
                artist=title[0].decode('iso-8859-15', 'replace'),
                year=info['EXTD'].replace("YEAR: ", ""),
                genre=info['DGENRE'],
            )

        self.name = title[1].decode('iso-8859-15', 'replace')
        event.log_event('cddb_info_retrieved', self, True)


class CDDevice(KeyedDevice):
    """
    represents a CD
    """

    class_autoconnect = True

    def __init__(self, dev):
        Device.__init__(self, dev)
        self.name = _("Audio Disc")
        self.dev = dev

    def _get_panel_type(self):
        import imp

        try:
            _cdguipanel = imp.load_source(
                "_cdguipanel", os.path.join(os.path.dirname(__file__), "_cdguipanel.py")
            )
            return _cdguipanel.CDPanel
        except Exception:
            logger.exception("Could not import cd gui panel")
            return 'flatplaylist'

    panel_type = property(_get_panel_type)

    def connect(self):
        cdpl = CDPlaylist(device=self.dev)
        self.playlists.append(cdpl)
        self.connected = True

    def disconnect(self):
        self.playlists = []
        self.connected = False
        CDDevice.destroy(self)


class UDisks2CdProvider(UDisksProvider):
    name = 'cd'
    PRIORITY = UDisksProvider.NORMAL

    def _get_num_tracks(self, obj, udisks):
        if obj.iface_type != 'org.freedesktop.UDisks2.Block':
            return

        try:
            drive = udisks.get_object_by_path(obj.props.Get('Drive'))
        except KeyError:
            return

        # Use number of audio tracks to identify supported media
        ntracks = drive.props.Get('OpticalNumAudioTracks')
        if ntracks > 0:
            return ntracks

    def get_priority(self, obj, udisks):
        ntracks = self._get_num_tracks(obj, udisks)
        if ntracks is not None:
            return self.PRIORITY

    def get_device(self, obj, udisks):
        device = obj.props.Get('Device', byte_arrays=True).decode('utf-8')
        device = device.strip('\0')
        return CDDevice(device)

    def on_device_changed(self, obj, udisks, device):
        if self._get_num_tracks(obj, udisks) is None:
            return 'remove'


# vim: et sts=4 sw=4



# Copyright (C) 2009-2010 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os.path

from xl import providers
from xl.nls import gettext as _
from xlgui import icons
from xlgui.preferences import widgets

name = _('Mini Mode')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "minimode_preferences.ui")
icons.MANAGER.add_icon_name_from_directory(
    'exaile-minimode', os.path.join(basedir, 'icons')
)
icon = 'exaile-minimode'


class AlwaysOnTopPreference(widgets.CheckPreference):
    name = 'plugin/minimode/always_on_top'
    default = True


class ShowInPanelPreference(widgets.CheckPreference):
    name = 'plugin/minimode/show_in_panel'
    default = False


class OnAllDesktopsPreference(widgets.CheckPreference):
    name = 'plugin/minimode/on_all_desktops'
    default = True


class ButtonInMainWindowPreference(widgets.CheckPreference):
    name = 'plugin/minimode/button_in_mainwindow'
    default = False


class DisplayWindowDecorationsPreference(widgets.CheckPreference):
    name = 'plugin/minimode/display_window_decorations'
    default = True


class WindowDecorationTypePreference(widgets.ComboPreference, widgets.CheckConditional):
    name = 'plugin/minimode/window_decoration_type'
    default = 'full'
    condition_preference_name = 'plugin/minimode/display_window_decorations'

    def __init__(self, preferences, widget):
        widgets.ComboPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class UseAlphaTransparencyPreference(widgets.CheckPreference):
    default = False
    name = 'plugin/minimode/use_alpha'


class TransparencyPreference(widgets.ScalePreference, widgets.CheckConditional):
    default = 0.3
    name = 'plugin/minimode/transparency'
    condition_preference_name = 'plugin/minimode/use_alpha'

    def __init__(self, preferences, widget):
        widgets.ScalePreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


class SelectedControlsPreference(widgets.SelectionListPreference):
    name = 'plugin/minimode/selected_controls'
    default = [
        'previous',
        'play_pause',
        'next',
        'playlist_button',
        'progress_bar',
        'restore',
    ]

    def __init__(self, preferences, widget):
        self.items = [
            self.Item(p.name, p.title, p.description, p.fixed)
            for p in providers.get('minimode-controls')
        ]
        widgets.SelectionListPreference.__init__(self, preferences, widget)


class TrackTitleFormatPreference(widgets.ComboEntryPreference):
    name = 'plugin/minimode/track_title_format'
    completion_items = {
        '$tracknumber': _('Track number'),
        '$title': _('Title'),
        '$artist': _('Artist'),
        '$composer': _('Composer'),
        '$album': _('Album'),
        '$__length': _('Length'),
        '$discnumber': _('Disc number'),
        '$__rating': _('Rating'),
        '$date': _('Date'),
        '$genre': _('Genre'),
        '$bitrate': _('Bitrate'),
        '$__loc': _('Location'),
        '$filename': _('Filename'),
        '$__playcount': _('Play count'),
        '$__last_played': _('Last played'),
        '$bpm': _('BPM'),
    }
    preset_items = [
        # TRANSLATORS: Mini mode track selector title preset
        _('$tracknumber - $title'),
        # TRANSLATORS: Mini mode track selector title preset
        _('$title by $artist'),
        # TRANSLATORS: Mini mode track selector title preset
        _('$title ($__length)'),
    ]
    default = _('$tracknumber - $title')



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import inspect
import locale
import logging

from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk

from xl import xdg
from xl.nls import gettext as _
from xlgui import icons
from . import (
    appearance,
    collection,
    cover,
    lyrics,
    playback,
    playlists,
    plugin,
    widgets,
)

logger = logging.getLogger(__name__)


class PreferencesDialog:
    """
    Preferences Dialog
    """

    PAGES = (playlists, appearance, playback, collection, cover, lyrics)
    PREFERENCES_DIALOG = None

    def __init__(self, parent, main):
        """
        Initializes the preferences dialog
        """
        self.main = main
        self.last_child = None
        self.last_page = None
        self.parent = parent
        self.fields = {}
        self.panes = {}
        self.builders = {}
        self.popup = None

        self.builder = Gtk.Builder()
        self.builder.set_translation_domain('exaile')
        self.builder.add_from_file(
            xdg.get_data_path('ui', 'preferences', 'preferences_dialog.ui')
        )
        self.builder.connect_signals(self)

        self.window = self.builder.get_object('PreferencesDialog')
        self.window.set_transient_for(parent)
        self.window.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.window.connect('delete-event', lambda *e: self.close())

        self.box = self.builder.get_object('preferences_box')

        self.tree = self.builder.get_object('preferences_tree')
        self.model = self.builder.get_object('model')

        title_cellrenderer = self.builder.get_object('title_cellrenderer')
        title_cellrenderer.props.ypad = 3

        self.default_icon = icons.MANAGER.pixbuf_from_icon_name(
            'document-properties', Gtk.IconSize.MENU
        )

        # sets up the default panes
        for page in self.PAGES:
            icon = self.default_icon

            if hasattr(page, 'icon'):
                if isinstance(page.icon, GdkPixbuf.Pixbuf):
                    icon = page.icon
                else:
                    icon = icons.MANAGER.pixbuf_from_icon_name(
                        page.icon, Gtk.IconSize.MENU
                    )

            self.model.append(None, [page, page.name, icon])

        # Use icon name to allow overrides
        plugin_icon = icons.MANAGER.pixbuf_from_icon_name(
            'extension', Gtk.IconSize.MENU
        )
        self.plug_root = self.model.append(None, [plugin, _('Plugins'), plugin_icon])

        self._load_plugin_pages()

        selection = self.tree.get_selection()
        selection.connect('changed', self.switch_pane)
        # Disallow selection on rows with no widget to show
        selection.set_select_function(
            (lambda sel, model, path, issel, dat: model[path][0] is not None), None
        )

        GLib.idle_add(selection.select_path, (0,))

    def _load_plugin_pages(self):
        self._clear_children(self.plug_root)
        plugin_pages = []
        plugin_manager = self.main.exaile.plugins

        for name in plugin_manager.enabled_plugins:
            plugin = plugin_manager.enabled_plugins[name]
            if hasattr(plugin, 'get_preferences_pane'):
                try:
                    plugin_pages.append(plugin.get_preferences_pane())
                except Exception:
                    logger.exception('Error loading preferences pane')

        plugin_pages.sort(key=lambda x: locale.strxfrm(x.name))

        for page in plugin_pages:
            icon = self.default_icon

            if hasattr(page, 'icon'):
                if isinstance(page.icon, GdkPixbuf.Pixbuf):
                    icon = page.icon
                else:
                    icon = icons.MANAGER.pixbuf_from_icon_name(
                        page.icon, Gtk.IconSize.MENU
                    )

            self.model.append(self.plug_root, [page, page.name, icon])

        GLib.idle_add(self.tree.expand_row, self.model.get_path(self.plug_root), False)

    def _clear_children(self, node):
        remove = []
        iter = self.model.iter_children(node)
        while iter:
            remove.append(iter)
            iter = self.model.iter_next(iter)

        for iter in remove:
            self.model.remove(iter)

    def on_close_button_clicked(self, widget):
        """
        Called when the user clicks 'ok'
        """
        self.close()

    def close(self):
        """
        Closes the preferences dialog
        """
        if hasattr(self.last_page, 'page_leave'):
            self.last_page.page_leave(self)
        self.window.hide()
        self.window.destroy()
        PreferencesDialog.PREFERENCES_DIALOG = None

    def switch_pane(self, selection):
        """
        Switches a pane
        """
        (model, iter) = selection.get_selected()
        if not iter:
            return
        page = self.model.get_value(iter, 0)
        if not page:
            return

        if self.last_child:
            self.box.remove(self.last_child)

        if self.last_page:
            if hasattr(self.last_page, 'page_leave'):
                self.last_page.page_leave(self)

        self.last_page = page

        child = self.panes.get(page)
        if not child:
            if hasattr(page, 'ui'):
                builder = Gtk.Builder()
                builder.add_from_file(page.ui)
            else:
                logger.error("No preference pane found")
                return

            child = builder.get_object('preferences_pane')
            init = getattr(page, 'init', None)
            if init:
                init(self, builder)
            self.panes[page] = child
            self.builders[page] = builder

        if page not in self.fields:
            self._populate_fields(page, self.builders[page])

        if hasattr(page, 'page_enter'):
            page.page_enter(self)

        child.unparent()
        self.box.pack_start(child, True, True, 0)
        self.last_child = child
        self.box.show_all()

    def _populate_fields(self, page, builder):
        """
        Populates field pages
        """
        self.fields[page] = []

        attributes = dir(page)

        for attr in attributes:
            try:
                klass = getattr(page, attr)

                if inspect.isclass(klass) and issubclass(klass, widgets.Preference):
                    widget = builder.get_object(klass.name)

                    if not widget:
                        logger.warning('Invalid preferences widget: %s', klass.name)
                        continue

                    if issubclass(klass, widgets.Conditional):
                        klass.condition_widget = builder.get_object(
                            klass.condition_preference_name
                        )
                    elif issubclass(klass, widgets.MultiConditional):
                        for name in klass.condition_preference_names:
                            klass.condition_widgets[name] = builder.get_object(name)

                    field = klass(self, widget)

                    label_widget = builder.get_object('label:%s' % klass.name)
                    if label_widget:
                        field.label_widget = label_widget

                    self.fields[page].append(field)
            except Exception:
                logger.exception('Broken preferences class: %s', attr)

    def run(self):
        """
        Runs the dialog
        """
        if PreferencesDialog.PREFERENCES_DIALOG:
            self = PreferencesDialog.PREFERENCES_DIALOG
            self.window.present()
        else:
            PreferencesDialog.PREFERENCES_DIALOG = self
            self.window.show_all()


# vim: et sts=4 sw=4
